"""
The __init__.py file for the stitches package.

The stitches package provides tools for stitching together climate model output
into a single, coherent dataset.
"""


from ._version import __version__
from .fx_match import match_neighborhood
from .fx_pangeo import fetch_nc, fetch_pangeo_table
from .fx_recipe import generate_gridded_recipe, make_recipe, permute_stitching_recipes
from .fx_stitch import gmat_stitching, gridded_stitching
from .generate_package_data import generate_pkg_data
from .install_pkgdata import install_package_data
from .make_matching_archive import make_matching_archive
from .make_pangeo_table import make_pangeo_comparison, make_pangeo_table
from .make_tas_archive import make_tas_archive
from .package_data import fetch_quickstarter_data

__all__ = [
    "match_neighborhood",
    "fetch_nc",
    "fetch_pangeo_table",
    "generate_gridded_recipe",
    "make_recipe",
    "permute_stitching_recipes",
    "gmat_stitching",
    "gridded_stitching",
    "generate_pkg_data",
    "install_package_data",
    "make_matching_archive",
    "make_pangeo_comparison",
    "make_pangeo_table",
    "make_tas_archive",
    "fetch_quickstarter_data",
    "__version__",
]



"""Module for working with the Pangeo database.

This module contains functions that are useful for interacting with the Pangeo data base.
See https://pangeo.io/index.html for more details on Pangeo.
"""

import fsspec
import intake
import xarray as xr
from tqdm import tqdm


def fetch_pangeo_table():
    """
    Fetch the Pangeo CMIP6 archive table of contents as a pandas DataFrame.

    Retrieve a copy of the Pangeo CMIP6 archive contents, which includes information
    about the available models, sources, experiments, ensembles, and more.

    :return: A pandas DataFrame with details on the datasets available for download from Pangeo.
    """
    # The url path that contains to the pangeo archive table of contents.
    url = "https://storage.googleapis.com/cmip6/pangeo-cmip6.json"
    out = intake.open_esm_datastore(url)

    return out.df


def fetch_nc(zstore: str):
    """
    Extract data for a single file from Pangeo.

    :param zstore: The location of the CMIP6 data file on Pangeo.
    :type zstore: str
    :return: An xarray Dataset containing CMIP6 data downloaded from Pangeo.
    """
    print(f"Fetching: {zstore}")

    # Function to update the progress bar
    def update_progress_bar(mapper, bar):
        keys = list(mapper.keys())

        for key in keys:
            _ = mapper[key]  # Trigger the actual read
            bar.update(1)

    # Create a file system mapper
    mapper = fsspec.get_mapper(zstore)

    # Initialize the progress bar
    with tqdm(total=len(mapper.keys()), desc="Downloading file components: ") as bar:
        update_progress_bar(mapper, bar)

    # Open the dataset
    ds = xr.open_zarr(mapper)
    ds.sortby("time")

    return ds



import unittest

import stitches
import stitches.install_pkgdata as sd


class TestInstallRawData(unittest.TestCase):
    """Tests for verifying the installation of raw data."""

    def test_instantiate(self):
        """Test instantiation of InstallPackageData with a fake data directory."""
        zen = sd.InstallPackageData(data_dir="fake")

        # Ensure default version is set
        self.assertEqual(str, type(zen.DEFAULT_VERSION))

        # Ensure URLs are present for current version
        self.assertTrue(stitches.__version__ in zen.DATA_VERSION_URLS)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np
import pandas as pd
import xarray as xr

from stitches.fx_pangeo import fetch_nc
from stitches.fx_stitch import (
    find_var_cols,
    find_zfiles,
    gmat_stitching,
    gridded_stitching,
    internal_stitch,
)
from stitches.fx_util import nrow


class TestStitch(unittest.TestCase):
    """
    Unit tests for stitching functions in the `stitches` package.

    This class provides a set of tests to ensure the correct functionality
    of the stitching functions, which are used to combine different climate
    model outputs into a single coherent dataset.
    """

    RUN = "ci"

    # This is an example recipe that will be used to test the stitching functions
    MY_RP = pd.DataFrame(
        data={
            "target_start_yr": [1850, 1859],
            "target_end_yr": [1858, 1867],
            "archive_experiment": ["historical", "historical"],
            "archive_variable": ["tas", "tas"],
            "archive_model": ["BCC-CSM2-MR", "BCC-CSM2-MR"],
            "archive_ensemble": ["r1i1p1f1", "r1i1p1f1"],
            "stitching_id": ["ssp245~r1i1p1f1~1", "ssp245~r1i1p1f1~1"],
            "archive_start_yr": [1859, 1886],
            "archive_end_yr": [1867, 1894],
            "tas_file": [
                "gs://cmip6/CMIP6/CMIP/BCC/BCC-CSM2-MR/historical/r1i1p1f1/Amon/tas/gn/v20181126/",
                "gs://cmip6/CMIP6/CMIP/BCC/BCC-CSM2-MR/historical/r1i1p1f1/Amon/tas/gn/v20181126/",
            ],
        }
    )

    def test_find_var_cols(self):
        """Test the `find_var_cols` function for identifying variable columns."""
        o = pd.DataFrame(data={"tas": [1, 2], "col2": [3, 4]})
        self.assertEqual(len(find_var_cols(o)), 0)

        o = pd.DataFrame(data={"tas_file": [1, 2], "col2": [3, 4]})
        self.assertEqual(find_var_cols(o), ["tas"])

        o = pd.DataFrame(data={"tas_file": [1, 2], "col2": [3, 4], "fake_file": [1, 2]})
        self.assertEqual(len(find_var_cols(o)), 2)

    def test_find_zfiles(self):
        """
        Test the `find_zfiles` function to ensure it correctly identifies zipped file paths.

        This test verifies that the `find_zfiles` function correctly identifies file paths
        for zipped files within a given DataFrame.
        """
        d = pd.DataFrame(data={"tas": [1, 2], "col2": [3, 4], "year": [1, 2]})
        self.assertEqual(len(find_zfiles(d)), 0)

        d = pd.DataFrame(
            data={
                "tas_file": ["file1.csv", "file2.csv"],
                "col2": [3, 4],
                "year": [1, 2],
            }
        )
        file_list = find_zfiles(d)
        self.assertEqual(type(file_list), np.ndarray)
        self.assertEqual(len(file_list), 2)

        d = pd.DataFrame(
            data={
                "tas_file": ["file1.csv", "file1.csv"],
                "col2": [3, 4],
                "year": [1, 2],
            }
        )
        file_list = find_zfiles(d)
        self.assertEqual(len(file_list), len(np.unique(file_list)))
        self.assertTrue(len(file_list) != nrow(d))

    def test_gmat_stitching(self):
        """
        Test the output returned by `gmat_stitching`.

        This test checks the type and structure of the output to ensure it meets expected formats.
        """

        out = gmat_stitching(self.MY_RP)

        self.assertEqual(type(out), pd.core.frame.DataFrame)

        time_steps = (
            max(self.MY_RP["target_end_yr"]) - min(self.MY_RP["target_start_yr"]) + 1
        )
        self.assertEqual(nrow(out), time_steps)

        # If the recipe is read in backwards, it shouldn't matter. The output should be the same.
        reverse = self.MY_RP.copy()
        reverse = reverse.iloc[::-1]
        out2 = gmat_stitching(reverse)
        self.assertEqual(out.shape, out2.shape)
        self.assertEqual(out["year"][0], out2["year"][0])
        self.assertEqual(out["value"][6], out["value"][6])

        # Manipulate the recipe, sometimes it will be fine other times it will throw an error.
        rp = self.MY_RP.copy()

        rp["tas_file"] = ["fake.nc", "fake.nc"]
        out = gmat_stitching(rp)
        self.assertEqual(type(out), pd.core.frame.DataFrame)

        # If the recpie is missing a column the stitching function should fail.
        with self.assertRaises(KeyError):
            gmat_stitching(rp.drop("tas_file"))

        with self.assertRaises(KeyError):
            gmat_stitching(rp.drop("target_start_yr"))

        with self.assertRaises(IndexError):
            rp["archive_model"] = ["fake", "fake"]
            gmat_stitching(rp)

    def test_gridded_related(self):
        """
        Test functions related to gridded data stitching.

        This test suite covers the functionality of gridded data stitching,
        ensuring that the output is consistent and errors are raised when
        expected.
        """
        if TestStitch.RUN == "ci":
            self.assertEqual(0, 0)
        else:
            # Set up the elements required for internal_stitch
            file_list = find_zfiles(self.MY_RP)
            data_list = list(map(fetch_nc, file_list))
            rslt = internal_stitch(self.MY_RP, data_list, file_list)

            time_steps = (
                12
                * (
                    max(self.MY_RP["target_end_yr"])
                    - min(self.MY_RP["target_start_yr"])
                )
                + 12
            )
            self.assertEqual(len(rslt["tas"]["time"]), time_steps)

            # Now do the stitching
            out = gridded_stitching(".", self.MY_RP)
            data = xr.open_dataset(out[0])
            time1 = data["tas"]["time"].values
            self.assertEqual(type(data), xr.core.dataset.Dataset)
            self.assertEqual(len(data["time"]), time_steps)
            os.remove(out[0])

            # If the recipe is read in backwards, it shouldn't matter the output should be the same.
            reverse = self.MY_RP.copy()
            reverse = reverse.iloc[::-1]
            out = gridded_stitching(".", reverse)
            data2 = xr.open_dataset(out[0])
            time2 = data2["tas"]["time"].values
            os.remove(out[0])
            self.assertEqual(max(time1 - time2), 0)

            # Manipulate the recipe, sometimes it will be fine other times it will throw an error.
            rp = self.MY_RP.copy()
            with self.assertRaises(TypeError):
                gridded_stitching("fake", rp)
            with self.assertRaises(KeyError):
                gridded_stitching(".", rp.drop("tas_file"))


if __name__ == "__main__":
    unittest.main()



"""Collection of helper functions for generating permutations of recipes and reformatting for stitching."""

import os
from importlib import resources

import numpy as np
import pandas as pd

import stitches.fx_match as match
import stitches.fx_util as util


def get_num_perms(matched_data):
    """
    Calculate the number of potential permutations from matched data.

    This function takes the results of `match_neighborhood(target, archive)` and
    determines the total number of potential permutations of the matches that cover
    the period 1850-2100 in the matched_data dataframe. It also provides a breakdown
    of how many matches are in each period of the target data.

    :param matched_data: The data output from match_neighborhood.
    :return: A list with two entries: the total number of potential permutations, and
             a dataframe with the breakdown of matches per period.
    """
    # Check inputs
    util.check_columns(
        matched_data,
        {
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
        },
    )

    dat = matched_data.drop_duplicates()
    dat_count = (
        dat.groupby(
            [
                "target_variable",
                "target_experiment",
                "target_ensemble",
                "target_model",
                "target_start_yr",
                "target_end_yr",
                "target_year",
                "target_fx",
                "target_dx",
            ]
        )
        .size()
        .reset_index(name="n_matches")
    )
    dat_count = dat_count.sort_values(["target_year"])

    dat_min = (
        dat_count.groupby(
            ["target_variable", "target_experiment", "target_ensemble", "target_model"]
        )["n_matches"]
        .min()
        .reset_index(name="minNumMatches")
    )
    dat_prod = (
        dat_count.groupby(
            ["target_variable", "target_experiment", "target_ensemble", "target_model"]
        )["n_matches"]
        .prod()
        .reset_index(name="totalNumPerms")
    )
    dat_count_merge = dat_min.merge(dat_prod)

    out = [dat_count_merge, dat_count]
    return out


def remove_duplicates(md, archive):
    """
    Ensure each archive point in a matched recipe is unique.

    When two target tgav windows in the trajectory match to the same archive window,
    the target window with the smaller Euclidean distance retains the match. The other
    target window is re-matched with its nearest-neighbor from a new archive, which
    excludes all previously matched points.

    :param md: A data frame with the results of matching for a single tgav recipe.
               This can be from match_neighborhood specifically returning NN, or from
               multiple matches permuted into new recipes and then split, with this
               function applied to each recipe.
    :param archive: A data frame consisting of the tas archive for re-matching.
    :return: A data frame with the same structure as the raw matched data, but with
             duplicate matches replaced.
    """
    if len(md["target_year"].unique()) < util.nrow(md):
        raise TypeError(
            "You have multiple matches to a single target year, this function can only accept a matched "
            "data frame of singular matches between target & archive data."
        )

    #  Initialize everything that gets updated on each iteration of the while loop:
    # 1. the data frame of matched_data -> make a copy of the argument md to initialize
    # 2. the data frame of duplicates is calculated for the first time.
    matched_data = md.copy()

    # Check to see if in the matched data frame if there are any repeated values.
    md_archive = matched_data[
        [
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
        ]
    ]
    duplicates = matched_data.merge(md_archive[md_archive.duplicated()], how="inner")

    # As long as duplicates exist, rematch the target windows with the larger
    # dist l2 to each archive chunk, add back in, iterate to be safe.
    # By matching on new_archive = archive - matches that were used in md,
    # we don't introduce new duplicates when we rematch. So the while loop is
    # probably over cautious but it does only execute one iteration.
    while util.nrow(duplicates) > 0:
        # within each iteration of checking duplicates,
        # pull out the one with smallest dist_l2 -
        # this is the one that gets to keep the archive match, and we use
        # as an index to work on the complement of (in case the same
        # archive point gets matched for more than 2 target years)
        grouped = duplicates.groupby(
            [
                "archive_experiment",
                "archive_variable",
                "archive_model",
                "archive_ensemble",
                "archive_start_yr",
                "archive_end_yr",
                "archive_year",
                "archive_fx",
                "archive_dx",
            ]
        )
        # Pick which of the target points will continue to be matched with the archive
        # pair.
        dat = []
        for name, group in grouped:
            min_value = min(group["dist_l2"])
            dat.append(group.loc[group["dist_l2"] == min_value])
        duplicates_min = pd.concat(dat)

        # target points contained in duplicates-duplicates_min
        # are the  ones that need a new archive match.
        filter_col = [col for col in duplicates if col.startswith("target_")]
        points_to_rematch = duplicates[filter_col].loc[
            (~duplicates["target_year"].isin(duplicates_min["target_year"]))
        ]
        new_names = list(
            map(lambda x: x.replace("target_", ""), points_to_rematch.columns)
        )
        points_to_rematch.columns = new_names

        # Because we know that none of the archive values can be reused in the match,
        # discard the ones already used (eg in matched_data)
        # from the updated archive that will be used in the rematching.
        cols = [col for col in matched_data if col.startswith("archive_")]
        rm_from_archive = matched_data[cols]
        new_names = list(
            map(lambda x: x.replace("archive_", ""), rm_from_archive.columns)
        )
        rm_from_archive.columns = new_names

        # Use our anti_join utility function to return the rows of archive that are
        # not in rm_from_archive
        new_archive = util.anti_join(
            archive,
            rm_from_archive,
            bycols=[
                "model",
                "experiment",
                "variable",
                "ensemble",
                "start_yr",
                "end_yr",
                "year",
                "fx",
                "dx",
            ],
        )

        # Find new matches for the data the target data that is missing the archive pair. Because we
        # are only interested in completing our singular recipe the tol must be 0.
        rematched = match.match_neighborhood(
            target_data=points_to_rematch, archive_data=new_archive, tol=0
        )

        # Now, we update our key data frames for the next iteration of the while loop:
        # 1. matched_data gets updated to be rematched + (previous matched_data minus the targets
        # that were rematched).
        # 2. duplicates gets recreated, checking for duplicates in our updated matched_data.

        # update matched_data:
        # first, drop the target windows that got rematched from the current matched_data:
        matched_data_minus_rematched_targ_years = matched_data.loc[
            ~(matched_data["target_year"].isin(rematched["target_year"]))
        ].copy()

        matched_data = (
            pd.concat([matched_data_minus_rematched_targ_years, rematched])
            .sort_values("target_year")
            .reset_index(drop=True)
        )

        # Identify duplicates in the updated matched_data for the next iteration of the while loop
        md_archive = matched_data[
            [
                "archive_experiment",
                "archive_variable",
                "archive_model",
                "archive_ensemble",
                "archive_start_yr",
                "archive_end_yr",
                "archive_year",
                "archive_fx",
                "archive_dx",
            ]
        ]
        duplicates = matched_data.merge(
            md_archive[md_archive.duplicated()], how="inner"
        )

        # Clean up for the next while loop iteration
        del (
            duplicates_min,
            points_to_rematch,
            rm_from_archive,
            rematched,
            matched_data_minus_rematched_targ_years,
        )

    return matched_data


def permute_stitching_recipes(
    N_matches: int, matched_data, archive, optional=None, testing: bool = False
):
    """
    Sample from `matched_data` to produce permutations of stitching recipes.

    This function samples from `matched_data` (the results of `match_neighborhood(target, archive, tol)`)
    to produce permutations of possible stitching recipes that will match the target data.

    :param N_matches: The maximum number of matches per target data.
    :type N_matches: int

    :param matched_data: Data output from `match_neighborhood`.

    :param archive: The archive data to use for re-matching duplicate points.

    :param optional: A previous output of this function that contains a list of already created recipes
                     to avoid re-making (this is not implemented).

    :param testing: When True, the behavior can be reliably replicated without setting global seeds.
                    Defaults to False.
    :type testing: bool

    :return: A data frame with the same structure as the raw matched data, with duplicate matches replaced.
    """
    # Check inputs
    util.check_columns(
        matched_data,
        {
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
        },
    )

    # Initialize quantities updated on every iteration of the while loop:
    # 1. A copy of the matched data
    # 2. perm_guide

    # Initialize matched_data_int for iteration through the while loop:
    # make a copy of the data to work with to be sure we don't touch original argument
    matched_data_int = matched_data.drop_duplicates().copy()

    # identifying how many target windows are in a trajectory we want to
    # create so that we know we have created a full trajectory with no
    # missing windows; basically a reference for us to us in checks.
    num_target_windows = util.nrow(matched_data_int["target_year"].unique())

    # Initialize perm_guide for iteration through the while loop.
    # the permutation guide is one of the factors that the while loop
    # will run checks on, must be initialized.
    # Perm_guide is basically a dataframe where each target window
    # lists the number of archive matches it has.
    num_perms = get_num_perms(matched_data_int)
    perm_guide = num_perms[1]

    # how many target trajectories are we matching to,
    # how many collapse-free ensemble members can each
    # target support, and order them according to that
    # for construction.
    targets = num_perms[0].sort_values(["minNumMatches"]).reset_index()
    # Add a column of a target  id name, differentiate between the different input
    # streams we are emulating.
    # We specifically emulate starting with the realization that can support
    # the fewest collapse-free generated realizations and work in increasing
    # order from there. We iterate over the different realizations to facilitate
    # checking for duplicates across generated realizations across target
    # realizations.
    targets["target_ordered_id"] = ["A" + str(x) for x in targets.index]

    if util.nrow(num_perms[0]["target_experiment"].unique()) > 1:
        raise TypeError(
            "Function permute_stitching_recipes should be applied to separate data frames "
            "for each target experiment of interest (multiple target ensemble members for "
            "a single target experiment is fine)."
        )

    # max number of permutations per target without repeating across generated
    # ensemble members.
    N_data_max = min(num_perms[0]["minNumMatches"])

    if N_matches > N_data_max:
        print(
            "More recipes requested than possible for at least one target trajectories, returning what can"
        )

    # Initialize the number of matches to either 0 or the input read from optional:
    if type(optional) is str:
        print("initialize to the read-in: has not been translated")
    else:
        recipe_collection = pd.DataFrame()

    # Loop over each target ensemble member, creating N_matches generated
    # realizations via a while loop before moving to the next target.
    for target_id in targets["target_ordered_id"].unique():
        # subset the target info, the target df contains meta information about the run we
        # and the number of permutations and such.
        target = targets.loc[targets["target_ordered_id"] == target_id].copy()

        # initialize a recipes data frame holder for each target, for
        # the while loop to iterate on
        recipes_col_by_target = pd.DataFrame()
        var_name = target["target_variable"].unique()[0]
        exp = target["target_experiment"].unique()[0]
        mod = target["target_model"].unique()[0]
        ens = target["target_ensemble"].unique()[0]

        # While the following conditions are met continue to generate new recipes.
        # 1. While we have fewer matches than requested for the target ensemble_member,
        #    keep going.
        # 2. Filter the perm_guide to just the target ensemble member in this loop and
        #    make sure there are at least num_target_windows of time windows: basically
        #    make sure there is at least one remaining archive match to draw from for
        #    each target window in this target ensemble. Note this means the perm_guide
        #    must be updated at the end of every while loop iteration.
        #
        # Initialize these conditions, so we enter the while loop, then update again at the
        # end of each iteration:
        if util.nrow(recipes_col_by_target) == 0:
            condition1 = True
        elif util.nrow(recipes_col_by_target["stitching_id"].unique()) < N_matches:
            condition1 = True
        else:
            condition1 = False

        perm_rows = util.nrow(
            perm_guide.loc[
                (perm_guide["target_variable"] == var_name)
                & (perm_guide["target_experiment"] == exp)
                & (perm_guide["target_model"] == mod)
                & (perm_guide["target_ensemble"] == ens)
            ]
            .copy()
            .drop_duplicates()
        )

        if perm_rows == num_target_windows:
            condition2 = True
        else:
            condition2 = False

        # And an integer index to initialize the count of stitched
        # trajectories for each target
        stitch_ind = 1

        # Run the while loop!
        while all([condition1, condition2]):
            # Group matched data for a single target by the chunks of the target information.
            # Right now a single target chunk may have multiple matches with archive points. The
            # next several steps of the while loop will create a one to one paring between the
            # target and archive data, then check to make sure that the pairing meets the requirements
            # for what we call a recipe.
            grouped_targets = []
            grouped_targets = (
                matched_data_int.loc[
                    (matched_data_int["target_variable"] == var_name)
                    & (matched_data_int["target_experiment"] == exp)
                    & (matched_data_int["target_model"] == mod)
                    & (matched_data_int["target_ensemble"] == ens)
                ]
                .copy()
                .groupby(
                    [
                        "target_variable",
                        "target_experiment",
                        "target_ensemble",
                        "target_model",
                        "target_start_yr",
                        "target_end_yr",
                    ]
                )
            )

            # For each target window group,
            # Randomly select one of the archive matches to use.
            # This creates one_one_match, a candidate recipe.
            one_one_match = []
            for name, group in grouped_targets:
                if testing:
                    one_one_match.append(group.sample(1, replace=False, random_state=1))
                else:
                    one_one_match.append(group.sample(1, replace=False))
            one_one_match = pd.concat(one_one_match)
            one_one_match = one_one_match.reset_index(drop=True).copy()

            # Before we can accept our candidate recipe, one_one_match,
            # we run it through a lot of tests.

            # Force one_one_match to meet our first condition,
            # that each archive data point in the recipe must be unique.
            # Then give it a stitching id
            new_recipe = []
            new_recipe = remove_duplicates(one_one_match, archive)
            stitching_id = exp + "~" + ens + "~" + str(stitch_ind)
            new_recipe["stitching_id"] = stitching_id
            new_recipe = new_recipe.reset_index(drop=True).copy()

            # Make sure the new recipe isn't missing any years:
            if ~new_recipe.shape[0] == num_target_windows:
                raise TypeError(
                    "problem: the new single recipe is missing years of data!"
                )
            #  Make sure that no changes were made to the target years.
            if (
                sum(
                    ~new_recipe["target_start_yr"].isin(
                        set(matched_data_int["target_start_yr"])
                    )
                )
                > 0
            ):
                raise TypeError("problem the new single recipe target years!")

            # Compare the new_recipe to the previously drawn recipes across all target
            # ensembles.
            # There is no collapse within each target ensemble because  we remove the constructed
            # new_recipe from the matched_data at the end of each iteration of the while loop -
            # The sampled points CAN'T be used again for the current target ensemble member
            # for loop iteration, or for any other target ensemble members. Meaning we
            # avoid envelope collapse when targeting multiple realizations (you don't have
            # realization 1 and realization 4 2070 getting matched to the same archive point.
            # The code below is checking to make sure that our new_recipe doesn't exist
            # in the saved recipe_collection. This shouldn't be possible with how we update
            # our matched_data_int on every loop, but just to be cautious, we check.
            # Again, the challenge is seeing if our entire sample has
            # been included in recipes before, not just a row or two.

            if util.nrow(recipe_collection) != 0:
                # If previous recipes exist, we must create a comparison
                # data frame that checks each existing recipe in recipe_collection
                # against new_recipe and record True/False
                #
                # Compare the new recipe with the existing collection of all recipes.
                cols_to_use = [
                    "target_variable",
                    "target_experiment",
                    "target_model",
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_variable",
                    "archive_model",
                    "archive_ensemble",
                    "archive_start_yr",
                    "archive_end_yr",
                ]
                grouped_collection = recipe_collection.groupby(["stitching_id"])
                comparison = []
                for name, group in grouped_collection:
                    df1 = (
                        group[cols_to_use]
                        .copy()
                        .reset_index(drop=True)
                        .sort_index(axis=1)
                    )
                    df2 = (
                        new_recipe[cols_to_use]
                        .copy()
                        .reset_index(drop=True)
                        .sort_index(axis=1)
                    )
                    comparison.append(df1.equals(df2))

                # end for loop
            # end if statement
            else:
                # Otherwise, this is the first recipe we've done at all, so we set comparison manually
                # so that the next if statement triggers just like it was appending a new recipe to an
                # existing list.
                comparison = [False]
            # end else

            # If the new_recipe is not unique (aka, any(comparison) == True), then
            # we don't want it and we don't want to do anything else in this iteration of
            # the while loop. We DON'T update the matched_points or conditions, so the
            # while loop is forced to re-run so that another random draw is done to create
            # a new candidate new_recipe.
            # We check for what we want: the new recipe is the first or all(comparison)==False.
            # In either case, we are safe to keep new_recipe and update all the data frames
            # for the next iteration of the while loop.
            if not all(comparison):
                # add new_recipe to the list of recipes for this target ensemble
                recipes_col_by_target = pd.concat([recipes_col_by_target, new_recipe])

                # And we remove it from the matched_points_int so the archive
                # values used in this new_recipe can't be used to construct
                # subsequent realizations for this target ensemble member.
                # This updated matched_data_int is used in each iteration
                # of the while loop. Since we are removing the constructed
                # new_recipe from the matched_data_int at the end of each
                # iteration of the while loop, the sample points can't be
                # randomly drawn again for the next generated trajectory
                # of the current target ensemble member for loop iteration.

                # Now each (target_window, archive_window) combination must
                # be removed from matched data for all target ensemble members,
                # not just the one we are currently operating on.
                # This ensures that we don't get collapse in the generated
                # envelope across target ensemble members (e.g you don't
                # have realization 1 and realization 4 2070 getting matched
                # to the same archive point).
                # Use an anti-join
                matched_data_int = util.anti_join(
                    matched_data_int,
                    new_recipe.drop(["stitching_id"], axis=1).copy(),
                    bycols=[
                        "target_year",
                        "target_start_yr",
                        "target_end_yr",
                        "archive_experiment",
                        "archive_variable",
                        "archive_model",
                        "archive_ensemble",
                        "archive_start_yr",
                        "archive_end_yr",
                        "archive_year",
                    ],
                ).copy()

                # update permutation count info with the revised matched data so
                # the while loop behaves - this makes sure that every target window
                # in the perm_guide actually has at least one matched archive point
                # available for draws .
                # That way, we don't try to construct a trajectory with fewer years
                # than the targets.
                num_perms = get_num_perms(matched_data_int)
                perm_guide = num_perms[1]

                # Use the updated perm_guide to update
                # the while loop conditions:

                # Condition 1:
                # If we haven't reached the N_matches goal for this target ensemble
                if util.nrow(recipes_col_by_target) == 0:
                    condition1 = True
                elif (
                    util.nrow(recipes_col_by_target["stitching_id"].unique())
                    < N_matches
                ):
                    condition1 = True
                else:
                    condition1 = False
                # end updating Condition 1

                # Condition 2:
                # make sure each target window in the updated perm guide has at least one archive match available
                # to draw on the next iteration.
                perm_rows = util.nrow(
                    perm_guide.loc[
                        (perm_guide["target_variable"] == var_name)
                        & (perm_guide["target_experiment"] == exp)
                        & (perm_guide["target_model"] == mod)
                        & (perm_guide["target_ensemble"] == ens)
                    ]
                    .copy()
                    .drop_duplicates()
                )

                if perm_rows == num_target_windows:
                    condition2 = True
                else:
                    condition2 = False
                # end updating condition 2

                # Add to the stitch_ind, to update the count of stitched
                # trajectories for each target ensemble member.
                stitch_ind += 1

            # end if statement
        # end the while loop for this target ensemble member

        # Add the collection of the recipes for each of the targets into single df.
        recipe_collection = (
            pd.concat([recipe_collection, recipes_col_by_target])
            .reset_index(drop=True)
            .copy()
        )

    # end the for loop over target ensemble members

    # do outputs
    out = recipe_collection.reset_index(drop=True).copy()
    return out[
        [
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
            "stitching_id",
        ]
    ]


def handle_transition_periods(rp):
    """
    Handle transition periods in the recipe data frame.

    This function processes the recipe data frame to ensure that there are no overlapping
    historical and future experiments during transition periods. It inserts extra periods
    to separate historical and future scenarios.

    :param rp: A data frame of the recipe.
    :return: A data frame of the recipe with separated historical/future experiments, ready to join
        with Pangeo information.
    """
    util.check_columns(
        rp,
        {
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
            "stitching_id",
        },
    )

    def internal_func(x):
        # First check to see the archive period spans the historical to future scenario
        transition_period = (
            2014 in range(x["archive_start_yr"], x["archive_end_yr"] + 1)
        ) & (2015 in range(x["archive_start_yr"], x["archive_end_yr"] + 1))

        if transition_period:
            target_yrs = list(range(x["target_start_yr"], x["target_end_yr"] + 1))
            archive_yrs = list(range(x["archive_start_yr"], x["archive_end_yr"] + 1))
            hist_cut_off = 2014  # the final complete year of the historical experiment

            historical_yrs = list(filter(lambda x: x <= hist_cut_off, archive_yrs))
            future_yrs = list(set(archive_yrs).difference(set(historical_yrs)))

            # This is the information that is constant between the historical and future periods.
            constant_info = x.loc[
                x.index.isin(
                    {
                        "archive_variable",
                        "archive_model",
                        "archive_ensemble",
                        "stitching_id",
                    }
                )
            ]

            # Construct the historical period information
            d = {
                "target_start_yr": min(target_yrs),
                "target_end_yr": target_yrs[len(historical_yrs) - 1],
                "archive_experiment": "historical",
                "archive_start_yr": min(historical_yrs),
                "archive_end_yr": max(historical_yrs),
            }
            ser = pd.Series(
                data=d,
                index=[
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_start_yr",
                    "archive_end_yr",
                ],
            )
            historical_period = pd.concat([ser, constant_info]).to_frame().transpose()

            # Check to make sure the lengths of time are correct
            targ_len = (
                historical_period["target_end_yr"].values
                - historical_period["target_start_yr"].values
            )
            arch_len = (
                historical_period["archive_end_yr"].values
                - historical_period["archive_start_yr"].values
            )
            if targ_len != arch_len:
                raise TypeError(
                    "Problem with the length of the historical archive & target years."
                )

            # Now construct the future period information
            d = {
                "target_start_yr": target_yrs[len(historical_yrs)],
                "target_end_yr": target_yrs[len(target_yrs) - 1],
                "archive_experiment": x["archive_experiment"],
                "archive_start_yr": min(future_yrs),
                "archive_end_yr": max(future_yrs),
            }
            ser = pd.Series(
                data=d,
                index=[
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_start_yr",
                    "archive_end_yr",
                ],
            )
            # future_period = ser.append(constant_info).to_frame().transpose()
            future_period = pd.concat([ser, constant_info]).to_frame().transpose()

            # Check to make sure the lengths of time are correct
            targ_len = (
                future_period["target_end_yr"].values
                - future_period["target_start_yr"].values
            )
            arch_len = (
                future_period["archive_end_yr"].values
                - future_period["archive_start_yr"].values
            )
            if not targ_len == arch_len:
                raise TypeError(
                    "Problem with the length of the historical archive & target years."
                )

            # Combine the period information
            out = pd.concat([historical_period, future_period]).reset_index(drop=True)
        else:
            out = x.to_frame().transpose().reset_index(drop=True)
            out = out[
                [
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_variable",
                    "archive_model",
                    "archive_ensemble",
                    "stitching_id",
                    "archive_start_yr",
                    "archive_end_yr",
                ]
            ]

        return out

    # Note that data frame returned might not be identical in shape to the
    # recipe read in because any periods that cover the historical period
    # will be split into two rows.
    ser = rp.apply(internal_func, axis=1)
    out = pd.concat(ser.values.tolist()).reset_index(drop=True)
    return out


def handle_final_period(rp):
    """
    Ensure all periods in a recipe have matching target and archive lengths.

    This function processes a recipe data frame to ensure that each period has the
    same length for both the target and archive. This is particularly important for
    the final period of target data, which may be shorter than expected. If the target
    window is matched to the final period of archive data, the final archive year is
    repeated to provide a sufficient number of matches.

    :param rp: A data frame of the recipe.
    :return: A recipe data frame with target and archive periods of equal length.
    """

    def internal_func(x):
        """
        Process a single row of the recipe dataframe.

        This function is applied to each row of the recipe dataframe. It ensures that
        the target and archive periods have the same length by adjusting the archive
        end year if necessary. It also formats the row into a dataframe with a specific
        column order.

        :param x: A pandas Series representing a row of the recipe dataframe.
        :return: A pandas DataFrame representing the processed row, with target and
                 archive periods of equal length and columns in a specified order.
        """
        len_target = x["target_end_yr"] - x["target_start_yr"]
        len_archive = x["archive_end_yr"] - x["archive_start_yr"]

        if len_target == len_archive:
            # No problem return the the row as is
            out = x.to_frame().transpose().reset_index(drop=True)
            out = out[
                [
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_variable",
                    "archive_model",
                    "archive_ensemble",
                    "stitching_id",
                    "archive_start_yr",
                    "archive_end_yr",
                ]
            ]

        elif len_target < len_archive:
            # Figure out how much shorter the target period is than the archive period.
            out = x.to_frame().transpose().reset_index(drop=True)
            out = out[
                [
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_variable",
                    "archive_model",
                    "archive_ensemble",
                    "stitching_id",
                    "archive_start_yr",
                    "archive_end_yr",
                ]
            ]
            out["archive_end_yr"] = out["archive_end_yr"] - 1
        else:
            # Figure out how much shorter the target period is than the archive period.
            out = x.to_frame().transpose().reset_index(drop=True)
            out = out[
                [
                    "target_start_yr",
                    "target_end_yr",
                    "archive_experiment",
                    "archive_variable",
                    "archive_model",
                    "archive_ensemble",
                    "stitching_id",
                    "archive_start_yr",
                    "archive_end_yr",
                ]
            ]
            out["archive_start_yr"] = out["archive_start_yr"] - 1

        return out

    ser = rp.apply(internal_func, axis=1)
    out = pd.concat(ser.values.tolist()).reset_index(drop=True)

    return out


def generate_gridded_recipe(messy_recipe, res: str = "mon"):
    """
    Create a recipe for the stitching process using a messy recipe.

    :param messy_recipe: A data frame generated by the permute_recipes function.
    :param res: The resolution of the recipe, either 'mon' for monthly or 'day' for daily.
    :type res: str
    :return: A data frame formatted as a recipe for stitching.
    """
    # Check inputs
    util.check_columns(
        messy_recipe,
        {
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
            "stitching_id",
        },
    )
    if res not in ["mon", "day"]:
        raise TypeError("generate_gridded_recipe: does not recognize the res input.")

    # Clean up the recipe
    dat = handle_transition_periods(messy_recipe)
    dat = handle_final_period(dat)

    # Make sure that if there are historical years of data being used assign
    # the experiment name to historical.
    dat.loc[dat["archive_end_yr"] <= 2014, "archive_experiment"] = "historical"

    # Now that we have the formatted recipe add the pangeo tas information!!
    ptable_path = resources.files("stitches") / "data" / "pangeo_table.csv"
    pangeo_table = pd.read_csv(ptable_path)

    if res == "mon":
        table = "Amon"
    else:
        table = "day"
    tas_meta_info = pangeo_table.loc[
        (pangeo_table["variable"] == "tas") & (pangeo_table["domain"] == table)
    ]
    tas_meta_info = tas_meta_info[
        ["model", "experiment", "ensemble", "variable", "zstore"]
    ]
    tas_meta_info = tas_meta_info.rename(
        columns={
            "model": "archive_model",
            "experiment": "archive_experiment",
            "ensemble": "archive_ensemble",
            "variable": "archive_variable",
        }
    )

    out = dat.merge(
        tas_meta_info,
        how="inner",
        on=[
            "archive_model",
            "archive_experiment",
            "archive_ensemble",
            "archive_variable",
        ],
    )
    out = out.reset_index(drop=True).copy()
    out = out.sort_values(["stitching_id", "target_start_yr", "target_end_yr"]).copy()
    out = out.reset_index(drop=True).copy()

    # Make sure that no information has been lost, if that is the case raise an error.
    if util.nrow(out) != util.nrow(dat):
        raise TypeError("Problem in generate_gridded_recipe, loosing data in recipe!")

    return out


def make_recipe(
    target_data,
    archive_data,
    N_matches: int,
    res: str = "mon",
    tol: float = 0.1,
    non_tas_variables: [str] = None,
    reproducible: bool = False,
):
    """
    Generate a stitching recipe from target and archive data.

    :param target_data: A pandas DataFrame of climate information to emulate.
    :param archive_data: A pandas DataFrame of temperature data to use as the archive to match on.
    :param N_matches: The maximum number of matches per target data.
    :param res: Resolution of the stitched data, either 'mon' or 'day'.
    :param tol: Tolerance used in the matching process, default is 0.1.
    :param non_tas_variables: List of variables other than tas to stitch together; defaults to None,
        which stitches tas only.
    :param reproducible: If True, ensures reproducible behavior by using the testing=True argument
        in permute_stitching_recipes(); defaults to False.

    :type N_matches: int
    :type res: str
    :type tol: float
    :type non_tas_variables: list[str]
    :type reproducible: bool

    :return: A pandas DataFrame of a formatted recipe.
    """
    # Check the inputs
    util.check_columns(
        target_data,
        {
            "experiment",
            "variable",
            "ensemble",
            "model",
            "start_yr",
            "end_yr",
            "year",
            "fx",
            "dx",
        },
    )
    util.check_columns(
        archive_data,
        {
            "experiment",
            "variable",
            "ensemble",
            "model",
            "start_yr",
            "end_yr",
            "year",
            "fx",
            "dx",
        },
    )
    if not type(N_matches) is int:
        raise TypeError("N_matches: must be an integer")
    if not type(tol) is float:
        raise TypeError("tol: must be a float")

    if target_data["unit"].unique() != archive_data["unit"].unique():
        raise TypeError("units of target and archive data do not match")

    # drop the units from each dataframe so matching functions don't need updates
    target_data = target_data[
        [
            "experiment",
            "variable",
            "ensemble",
            "model",
            "start_yr",
            "end_yr",
            "year",
            "fx",
            "dx",
        ]
    ].copy()
    archive_data = archive_data[
        [
            "experiment",
            "variable",
            "ensemble",
            "model",
            "start_yr",
            "end_yr",
            "year",
            "fx",
            "dx",
        ]
    ].copy()

    # If there are non tas variables to be stitched, subset the archive to limit
    # the coverage to only the entries with the complete coverage.
    if type(non_tas_variables) == list:
        if res not in ["mon", "day"]:
            raise TypeError("does not recognize the res input.")
        if "tas" in non_tas_variables:
            raise TypeError("non_tas_variables: cannot contain tas.")

        data_directory = resources.files("stitches") / "data"
        pt_path = os.path.join(data_directory, "pangeo_table.csv")
        pangeo_table = pd.read_csv(pt_path)

        var_list = pangeo_table["variable"].unique()
        if not set(non_tas_variables) <= set(var_list):
            raise TypeError(
                "One or more of the variables are not found in Pangeo table."
            )

        # Subset the pangeo table so that it contains the resolution & variables of data of interest.
        non_tas_variables.append("tas")
        pt_subset = pangeo_table.loc[
            (
                pangeo_table["domain"].str.contains(res)
                & pangeo_table["variable"].isin(non_tas_variables)
                & ~pangeo_table["zstore"].str.contains("AerChemMIP")
            )
        ].copy()

        # Format the variable column to the "file" so that we can
        pt_subset["variable"] = pt_subset["variable"] + "_file"
        wide_df = pt_subset.pivot(
            index=["model", "ensemble", "experiment"],
            columns="variable",
            values="zstore",
        )
        wide_df.reset_index(inplace=True)
        wide_df.columns.name = None
        wide_df = wide_df.dropna()

        # Select the archive entries for the model, ensemble, experiment to keep, there are the entries
        # that also have complete coverage for the variables listed in the non tas variable list.
        to_keep = wide_df.loc[
            :, wide_df.columns.isin(["model", "ensemble", "experiment"])
        ].drop_duplicates()
        archive_data = archive_data.merge(
            to_keep, on=["model", "ensemble", "experiment"], how="inner"
        ).copy()

    # Match the archive & target data together.
    match_df = match.match_neighborhood(target_data, archive_data, tol=tol)

    if reproducible:
        unformatted_recipe = permute_stitching_recipes(
            N_matches=N_matches,
            matched_data=match_df,
            archive=archive_data,
            testing=True,
        )
    else:
        unformatted_recipe = permute_stitching_recipes(
            N_matches=N_matches,
            matched_data=match_df,
            archive=archive_data,
            testing=False,
        )

    # Format the recipe into the dataframe that can be used by the stitching functions.
    recipe = generate_gridded_recipe(unformatted_recipe, res=res)
    recipe.columns = [
        "target_start_yr",
        "target_end_yr",
        "archive_experiment",
        "archive_variable",
        "archive_model",
        "archive_ensemble",
        "stitching_id",
        "archive_start_yr",
        "archive_end_yr",
        "tas_file",
    ]

    # If there are non tas variables add the non tas variables to the formatted recpie.
    if type(non_tas_variables) == list:
        to_join = wide_df.loc[
            :, ~wide_df.columns.isin(["model", "ensemble", "experiment"])
        ]
        out = pd.merge(recipe, to_join, on="tas_file")

    else:
        out = recipe.copy()

    out = (
        out.sort_values(by=["stitching_id", "target_start_yr"])
        .reset_index(drop=True)
        .copy()
    )
    return out



"""
The `fx_stitch` module is responsible for stitching together climate model outputs.

It creates a continuous time series that can be used for climate analysis and emulation.

"""

import os
from importlib import resources

import numpy as np
import pandas as pd
import xarray as xr

import stitches.fx_data as data
import stitches.fx_pangeo as pangeo
import stitches.fx_util as util


def find_zfiles(rp):
    """
    Determine which CMIP files must be downloaded from Pangeo.

    :param rp: Data frame of the recipes.
    :return: Numpy ndarray of the gs:// files to pull from Pangeo.
    """
    # Figure out which columns contain the string file
    flat_list = rp.filter(regex="file", axis=1).values.flatten()
    unique_list = np.unique(flat_list)
    return unique_list


def find_var_cols(x):
    """
    Determine the variables to be downloaded.

    :param x: pandas DataFrame of the stitches recipe.
    :return: List of variables to be written to the NetCDF files.
    """
    # Parse out the variable name so that we can use it
    # to label the final output.
    set = x.filter(regex="file").columns.tolist()

    out = []
    for text in set:
        new = text.replace("_file", "")
        out.append(new)
    return out


def get_netcdf_values(i, dl, rp, fl, name):
    """
    Extract archive values from a list of downloaded CMIP data.

    :param i: Index of the row in the recipe data frame.
    :param dl: List of xarray datasets containing CMIP files.
    :param rp: DataFrame of the recipe.
    :param fl: List of CMIP file paths.
    :param name: Name of the variable file to process.
    :return: A slice of xarray data (unsure about the technical term).
    """
    file = rp[name][i]
    start_yr = rp["archive_start_yr"][i]
    end_yr = rp["archive_end_yr"][i]

    # Figure out which index level we are on and then get the
    # xarray from the list.
    index = int(np.where(fl == file)[0])
    extracted = dl[index].sortby("time")
    v = name.replace("_file", "")

    # Have to have special time handler
    times = extracted.indexes["time"]

    if type(times) in [
        xr.coding.cftimeindex.CFTimeIndex,
        pd.core.indexes.datetimes.DatetimeIndex,
    ]:
        yrs = extracted.indexes[
            "time"
        ].year  # pull out the year information from the time index
        flags = list(map(lambda x: x in range(start_yr, end_yr + 1), yrs))
        to_keep = times[flags]
    else:
        raise TypeError("Unsupported time type.")
    dat = extracted.sel(time=to_keep)[v].values.copy()

    if (times.freq == "D") | (times.freq == "day"):
        expected_times = pd.date_range(
            start=str(start_yr) + "-01-01", end=str(end_yr) + "-12-31", freq="D"
        )
        if times.calendar == "noleap":
            expected_len = len(
                expected_times[
                    ~((expected_times.month == 2) & (expected_times.day == 29))
                ]
            )
    else:
        expected_len = len(
            pd.date_range(
                start=str(start_yr) + "-01-01", end=str(end_yr) + "-12-31", freq="M"
            )
        )

    assert len(dat) == expected_len, (
        "Not enough data in " + file + "for period " + str(start_yr) + "-" + str(end_yr)
    )

    return dat


def get_var_info(rp, dl, fl, name):
    """
    Extract the CMIP variable attribute information.

    :param rp: Data frame of the recipes.
    :param dl: List of the data files.
    :param fl: List of the data file names.
    :param name: String of the column containing the variable file name from rp.
    :return: Pandas dataframe of the variable meta data.
    """
    util.check_columns(rp, {name})
    file = rp[name][0]
    index = int(np.where(fl == file)[0])
    extracted = dl[index]

    attrs = data.get_ds_meta(extracted)
    if attrs.frequency.values != "mon":
        attrs["calendar"] = extracted.indexes["time"].calendar

    return attrs


def get_atts(rp, dl, fl, name):
    """
    Extract the CMIP variable attribute information.

    :param rp: Data frame of the recipes.
    :param dl: List of the data files.
    :param fl: List of the data file names.
    :param name: String of the column containing the variable file name from rp.
    :return: Dict object containing the CMIP variable information.
    """
    file = rp[name][0]
    index = int(np.where(fl == file)[0])
    extracted = dl[index]
    v = name.replace("_file", "")

    out = extracted[v].attrs.copy()

    return out


def internal_stitch(rp, dl, fl):
    """
    Stitch a single recipe into netCDF outputs.

    :param dl: List of xarray CMIP files.
    :param rp: DataFrame of the recipe.
    :param fl: List of the CMIP file names.
    :return: List of the data arrays for the stitched products of the different variables.
    """
    rp = rp.sort_values(by=["stitching_id", "target_start_yr"]).copy()
    rp.reset_index(drop=True, inplace=True)
    variables = find_var_cols(rp)
    out = []

    # For each of the of the variables stitch the
    # data together.
    for v in variables:
        # Get the information about the variable that is going to be stitched together.
        col = v + "_file"
        var_info = get_var_info(rp, dl, fl, col)

        # For each of time slices extract the data & concatenate together.
        gridded_data = get_netcdf_values(i=0, dl=dl, rp=rp, fl=fl, name=col)

        # Now add the other time slices.
        for i in range(1, len(rp)):
            new_vals = get_netcdf_values(i=i, dl=dl, rp=rp, fl=fl, name=col)
            gridded_data = np.concatenate((gridded_data, new_vals), axis=0)

        # Note that the pd.date_range call need the date/month defined otherwise it will
        # truncate the year from start of first year to start of end year which is not
        # what we want. We want the full final year to be included in the times series.
        start = str(min(rp["target_start_yr"]))
        end = str(max(rp["target_end_yr"]))

        if var_info["frequency"][0].lower() == "mon":
            freq = "M"
        elif var_info["frequency"][0].lower() == "day":
            freq = "D"
        else:
            raise TypeError("Unsupported frequency.")

        times = pd.date_range(start=start + "-01-01", end=end + "-12-31", freq=freq)

        # Again, some ESMs stop in 2099 instead of 2100 - so we just drop the
        # last year of gridded_data when that is the case.
        # TODO this will need something extra/different for daily data; maybe just
        # a simple len(times)==len(gridded_data)-12 : len(times) == len(gridded_data)-(nDaysInYear)
        # with correct parentheses would do it
        if (max(rp["target_end_yr"]) == 2099) & (
            len(times) == (len(gridded_data) - 12)
        ):
            gridded_data = gridded_data[0 : len(times), 0:, 0:].copy()

        if freq == "D":
            if (var_info["calendar"][0].lower() == "noleap") & (freq == "D"):
                times = times[~((times.month == 2) & (times.day == 29))]

        assert len(gridded_data) == len(times), "Problem with the length of time"

        # Extract the lat and lon information that will be used to structure the
        # empty netcdf file. Make sure to copy all of the information including
        # the attributes!
        lat = dl[0].lat.copy()
        lon = dl[0].lon.copy()

        rslt = xr.Dataset(
            {
                v: xr.DataArray(
                    gridded_data,
                    coords=[times, lat, lon],
                    dims=["time", "lat", "lon"],
                    attrs={
                        "units": var_info["units"][0],
                        "variable": var_info["variable"][0],
                        "experiment": var_info["experiment"][0],
                        "ensemble": var_info["ensemble"][0],
                        "model": var_info["model"][0],
                        "stitching_id": rp["stitching_id"].unique()[0],
                    },
                )
            }
        )

        out.append(rslt)

    out_dict = dict(zip(variables, out))

    return out_dict


def gridded_stitching(out_dir: str, rp):
    """
    Stitch the gridded NetCDFs for variables contained in the recipe file and save them.

    :param out_dir: Directory location where to write the NetCDF files.
    :type out_dir: str
    :param rp: DataFrame of the recipe including variables to stitch.
    :return: List of the NetCDF file paths.
    """
    flag = os.path.isdir(out_dir)
    if not flag:
        raise TypeError("The output directory does not exist.")

    # Check inputs.
    util.check_columns(
        rp,
        {
            "target_start_yr",
            "target_end_yr",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "stitching_id",
            "archive_start_yr",
            "archive_end_yr",
        },
    )

    rp = (
        rp.sort_values(by=["stitching_id", "target_start_yr"])
        .reset_index(drop=True)
        .copy()
    )

    # Determine which variables will be downloaded.
    variables = find_var_cols(rp)
    if not (len(variables) >= 1):
        raise KeyError("No variables were found to be processed.")

    # Determine which files need to be downloaded from pangeo.
    file_list = find_zfiles(rp)

    # Make sure that all of the files are available to download from pangeo.
    # Note that this might be excessively cautious but this is an issue we have run into in
    # the past.
    print("Validating request availability with Pangeo archive contents...")
    avail = pangeo.fetch_pangeo_table()
    flag = all(item in list(avail["zstore"]) for item in list(file_list))
    if not flag:
        raise KeyError("Trying to request a zstore file that does not exist.")

    # Download all of the data from pangeo.
    data_list = list(map(pangeo.fetch_nc, file_list))
    print(f"Compiled downloads into {len(data_list)} datasets.")

    # For each of the stitching recipes go through and stitch a recipe.
    for single_id in rp["stitching_id"].unique():
        # initialize f to be empty just to be safe now that we've added a
        # try...except approach. It's technically possible the first id
        # tried will fail and the function will try to return a non-existent f.
        f = []

        try:
            _model = rp.archive_model.unique()
            _var = rp.archive_variable.unique()
            action_msg = (
                f"Stitching gridded netcdf for: {_model} {_var} {single_id} ..."
            )
            print(action_msg)

            # Do the stitching!
            # ** this can be a slow step and prone to errors
            single_rp = rp.loc[rp["stitching_id"] == single_id].copy()

            rslt = internal_stitch(rp=single_rp, dl=data_list, fl=file_list)

            # Print the files out at netcdf files
            f = []
            for i in rslt.keys():
                ds = rslt[i]
                ds = ds.sortby("time").copy()
                recipe_location = (
                    out_dir
                    + "/"
                    + "stitched_"
                    + ds[i].attrs["model"]
                    + "_"
                    + ds[i].attrs["variable"]
                    + "_"
                    + single_id
                    + "_recipe.csv"
                )
                ds[i].attrs["recipe_location"] = recipe_location
                single_rp.to_csv(recipe_location, index=False)
                fname = (
                    out_dir
                    + "/"
                    + "stitched_"
                    + ds[i].attrs["model"]
                    + "_"
                    + ds[i].attrs["variable"]
                    + "_"
                    + single_id
                    + ".nc"
                )
                ds.to_netcdf(fname)
                f.append(fname)
            # end For loop over rslt keys
        # end try

        except KeyError:
            print(
                "Stitching gridded netcdf for: "
                + rp.archive_model.unique()
                + " "
                + rp.archive_variable.unique()
                + " "
                + single_id
                + " failed. Skipping. Error thrown within gridded_stitching fxn."
            )
        # end except
    # end for loop over single_id

    return f


def gmat_internal_stitch(row, data):
    """
    Select data from a tas archive based on a single row in a recipe data frame.

    This function is used to iterate over an entire recipe to do the stitching.

    :param row: A row entry of a fully formatted recipe as a pandas Series.
    :param data: A DataFrame containing the tas values to be stitched together.
    :return: A DataFrame of tas values.
    """
    years = list(range(int(row["target_start_yr"]), int(row["target_end_yr"]) + 1))
    select_years = list(
        range(int(row["archive_start_yr"]), int(row["archive_end_yr"]) + 1)
    )

    selected_data = data.loc[
        (data["experiment"] == row["archive_experiment"])
        & (data["year"].isin(select_years))
        & (data["ensemble"] == row["archive_ensemble"])
    ]

    # some models stop at 2099 instead of 2100 - so there is a mismatch
    # between len(years) and selected data but not a fatal one.
    # Write a very specific if statement to catch this & just chop the extra year
    # off the end of selected_data.
    if (len(years) == (util.nrow(selected_data) - 1)) & (max(years) == 2099):
        selected_data = selected_data.iloc[0 : len(years),].copy()

    if len(years) != util.nrow(selected_data):
        raise TypeError("Trouble with selecting the tas data.")

    new_vals = selected_data["value"]
    d = {"year": years, "value": new_vals}
    df = pd.DataFrame(data=d)
    df["variable"] = "tas"

    return df


def gmat_stitching(rp):
    """
    Stitch together a time series of global tas data based on a recipe data frame.

    :param rp: A fully formatted recipe data frame as a pandas DataFrame.
    :return: A pandas DataFrame of stitched together tas data.
    """
    # Check inputs.
    util.check_columns(
        rp,
        {
            "target_start_yr",
            "target_end_yr",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "stitching_id",
            "archive_start_yr",
            "archive_end_yr",
            "tas_file",
        },
    )

    rp = (
        rp.sort_values(by=["stitching_id", "target_start_yr"])
        .reset_index(drop=True)
        .copy()
    )

    # One the assumptions of this function is that it only works with tas, so
    # we can safely add tas as the variable column.
    rp["variable"] = "tas"
    out = []
    for name, match in rp.groupby("stitching_id"):
        # Reset the index in the match data frame so that we can use a for loop
        # to iterate through match data frame an apply the gmat_internal_stitch.
        match = match.reset_index(drop=True)

        # Find the tas data to be stitched together.
        dir_path = resources.files("stitches") / "data" / "tas-data"
        all_files = util.list_files(dir_path)

        # Load the tas data for a particular model.
        model = match["archive_model"].unique()[0]
        csv_to_load = [
            file
            for file in all_files
            if (model in file) and (os.path.basename(file)[0] != ".")
        ][0]

        data = pd.read_csv(csv_to_load)

        # Format the data so that if we have historical years in the future scenarios
        # then that experiment is relabeled as "historical".
        fut_exps = [
            "ssp245",
            "ssp126",
            "ssp585",
            "ssp119",
            "ssp370",
            "ssp434",
            "ssp534-over",
            "ssp460",
        ]
        nonssp_data = data.loc[~data["experiment"].isin(fut_exps)]
        fut_data = data.loc[
            (data["experiment"].isin(fut_exps)) & (data["year"] > 2014)
        ].copy()
        hist_data = data.loc[
            (data["experiment"].isin(fut_exps)) & (data["year"] <= 2014)
        ].copy()
        hist_data["experiment"] = "historical"
        tas_data = (
            pd.concat([nonssp_data, fut_data, hist_data])[
                ["variable", "experiment", "ensemble", "model", "year", "value"]
            ]
            .drop_duplicates()
            .reset_index(drop=True)
        )
        # Stitch the data together based on the matched recipes.
        dat = []
        for i in match.index:
            row = match.iloc[i, :]
            dat.append(gmat_internal_stitch(row, tas_data))
        dat = pd.concat(dat)

        # Add the stitching id column to the data frame.
        dat["stitching_id"] = name

        # Add the data to the out list
        out.append(dat)

    # Format the list of data frames into a single data frame.
    final_output = pd.concat(out)
    final_output = final_output.reset_index(drop=True).copy()
    final_output = final_output.sort_values(["stitching_id", "year"]).copy()
    final_output = final_output.reset_index(drop=True).copy()

    return final_output



"""
This module defines functions used to create the archive for the matching process.

It calculates the rate of change (dx) and median value (fx) for the temperature anomaly time series.
"""

from importlib import resources

import numpy as np
import pandas as pd

# Import packages
import stitches.fx_processing as prep
import stitches.fx_util as util


def make_matching_archive(
    smoothing_window: int = 9, chunk_window: int = 9, add_staggered: bool = False
):
    """
    Create an archive of rate of change (dx) and mean (fx) values.

    This function processes the CMIP6 archive to produce values used in the
    matching portion of the stitching pipeline.

    :param smoothing_window: The size of the smoothing window to be applied to the time series.
                             Defaults to 9.
    :type smoothing_window: int

    :param chunk_window: The size of the chunks of data to summarize with dx & fx.
                         Defaults to 9.
    :type chunk_window: int

    :param add_staggered: If True, staggered windows will be added to the archive.
                          Defaults to False.
    :type add_staggered: bool

    :return: The file location of the matching archive.
    """
    # Start by loading all of the tas files.
    raw_data = util.load_data_files("data/tas-data")
    raw_data = raw_data[raw_data["year"] <= 2100].copy()

    # Smooth the anomalies, get the running mean of each time series.
    # Each year in the original time series is retained, and the running mean
    # recorded for each year is the mean centered on that year across
    # smoothing_window number of years.
    smoothed_data = prep.calculate_rolling_mean(raw_data, smoothing_window)

    # For each group in the data set go through, chunk and extract the fx and dx␣ values.
    # For now we have to do this with a for loop, to process each model/experiment/ensemble/variable
    # individually.
    # The key function prepfor aring these chunks is prep.chunk_ts
    data = smoothed_data[
        ["model", "experiment", "ensemble", "year", "variable", "value"]
    ]
    data = data.reset_index(drop=True).copy()
    group_by = ["model", "experiment", "ensemble", "variable", "unit"]
    out = []
    for key, d in data.groupby(group_by):
        dat = d.reset_index(drop=True).copy()

        # if this data set doesn't have at least chunk_window worth of years,
        # just print a message that it isn't getting processed into chunks.
        # (it doesn't make sense to create a 9 year chunk window from 6
        # years of data, and it causes issues when we want to add the staggered
        if dat["year"].nunique() < chunk_window:
            mod = dat.model.unique()[0]
            exp = dat.experiment.unique()[0]
            ens = dat.ensemble.unique()[0]
            print(
                mod
                + "  "
                + exp
                + "  "
                + ens
                + "  has fewer than chunk_window="
                + str(chunk_window)
                + " years in its time series. Skipping"
            )
        else:
            dd = prep.chunk_ts(df=dat, n=chunk_window)
            rslt = prep.get_chunk_info(dd)
            rslt["unit"] = (np.repeat(d["unit"].unique(), len(rslt))).copy()
            out.append(rslt)
        # end if-else
    # end of the for loop

    # if adding staggered windows, do it now.
    if add_staggered:
        # for each offset, do the prep and append.
        for offset in range(1, chunk_window):
            for key, d in data.groupby(group_by):
                dat = d.reset_index(drop=True).copy()

                # if this data set doesn't have at least chunk_window worth of years,
                # just print a message that it isn't getting processed into chunks.
                if dat["year"].nunique() < chunk_window:
                    mod = dat.model.unique()[0]
                    exp = dat.experiment.unique()[0]
                    ens = dat.ensemble.unique()[0]
                    print(
                        mod
                        + "  "
                        + exp
                        + "  "
                        + ens
                        + "  has fewer than chunk_window="
                        + str(chunk_window)
                        + " years in its time series. Skipping"
                    )
                else:
                    dd = prep.chunk_ts(df=dat, n=chunk_window, base_chunk=offset)
                    rslt = prep.get_chunk_info(dd)
                    rslt["unit"] = (np.repeat(d["unit"].unique(), len(rslt))).copy()
                    out.append(rslt)
                # end if-else
            # end of the for loop over (model-experiment-ensemble-variable) combos
        # end for loop over base_chunk offsets
    # end if statement for adding staggered chunks

    #  concatenate results into a single data frame.
    data = pd.concat(out).reset_index(drop=True)

    outdir_path = resources.files("stitches") / "data"
    if add_staggered:
        ofile = outdir_path + "/matching_archive_staggered.csv"
        # for the staggered archive, because we've added so many points to
        # the archive, we keep only the points based on chunk_window years of
        # data (basically cutting out head and tail points, which we can afford
        # to lose).
        # This way, we don't have to change any of our stitching functions to handle
        # those cases when we use the staggered archive
        data = data[(data["end_yr"] - data["start_yr"] + 1 == chunk_window)].copy()
        data.to_csv(ofile, index=False)
    else:
        ofile = outdir_path + "/matching_archive.csv"
        data.to_csv(ofile, index=False)

    return ofile



"""Module for working with temperature data files."""

import numpy as np
import pandas as pd


def get_lat_name(ds):
    """Get the name for the latitude values in an xarray dataset.

    This function searches for latitude coordinates in the dataset,
    which could be named either 'lat' or 'latitude'.

    :param ds: The dataset from which to retrieve the latitude coordinate name.
    :type ds: xarray.Dataset
    :returns: The name of the latitude variable.
    :rtype: str
    :raises RuntimeError: If no latitude coordinate is found in the dataset.
    """
    for lat_name in ["lat", "latitude"]:
        if lat_name in ds.coords:
            return lat_name
    raise RuntimeError("Couldn't find a latitude coordinate")


def global_mean(ds):
    """
    Calculate the weighted global mean for a variable in an xarray dataset.

    :param ds: The xarray dataset of CMIP data.
    :type ds: xarray.Dataset
    :returns: The xarray dataset of the weighted global mean.
    :rtype: xarray.Dataset
    """
    lat = ds[get_lat_name(ds)]
    weight = np.cos(np.deg2rad(lat))
    weight /= weight.mean()
    other_dims = set(ds.dims) - {"time"}
    return (ds * weight).mean(other_dims)


def get_ds_meta(ds):
    """
    Get the metadata information from an xarray dataset.

    :param ds: xarray dataset of CMIP data.
    :return: pandas DataFrame of MIP information.
    """
    v = ds.variable_id

    data = [
        {
            "variable": v,
            "experiment": ds.experiment_id,
            "units": ds[v].attrs["units"],
            "frequency": ds.attrs["frequency"],
            "ensemble": ds.attrs["variant_label"],
            "model": ds.source_id,
        }
    ]
    df = pd.DataFrame(data)

    return df



"""This module contains the InstallPackageData class for downloading and unpacking example data from Zenodo."""

import importlib
import os
import shutil
import tempfile
import zipfile
from importlib import resources
from io import BytesIO as BytesIO

import requests
from tqdm import tqdm

from ._version import __version__


class InstallPackageData:
    """
    Download and unpack example data minted on Zenodo that matches the current installed stitches distribution.

    This class handles the retrieval and organization of example data for the stitches package.
    It ensures that the data version corresponds to the version of the installed stitches package.

    :param data_dir: Optional. Full path to the directory where you wish to store the data.
                     If not specified, the data will be installed in the data directory of the package.
    :type data_dir: str
    """

    # URL for DOI minted example data hosted on Zenodo
    DATA_VERSION_URLS = {
        "0.9.1": "https://zenodo.org/record/7181977/files/data.zip?download=1",
        "0.10.0": "https://zenodo.org/record/7799725/files/data.zip?download=1",
        "0.11.0": "https://zenodo.org/records/8367628/files/data.zip?download=1",
        "0.12.0": "https://zenodo.org/records/8367628/files/data.zip?download=1",
        "0.12.1": "https://zenodo.org/records/8367628/files/data.zip?download=1",
    }

    DEFAULT_VERSION = "https://zenodo.org/record/7167526/files/data.zip?download=1"

    def __init__(self, data_dir=None):
        """
        Initialize the InstallPackageData class.

        :param data_dir: The directory where the data will be stored. If None, the data
                         will be installed in the package's data directory.
        :type data_dir: str, optional
        """
        self.data_dir = data_dir

    def fetch_zenodo(self):
        """Download and unpack the Zenodo minted data for the current stitches distribution."""
        # full path to the stitches root directory where the example dir will be stored
        if self.data_dir is None:
            data_directory = resources.files("stitches") / "data"
        else:
            data_directory = self.data_dir

        # build needed subdirectories if they do not already exist
        tas_data_path = os.path.join(data_directory, "tas-data")
        temp_data_path = os.path.join(data_directory, "temp-data")
        if not os.path.exists(tas_data_path):
            os.mkdir(tas_data_path)
        if not os.path.exists(temp_data_path):
            os.mkdir(temp_data_path)

        # get the current version of stitches that is installed
        current_version = __version__

        try:
            data_link = InstallPackageData.DATA_VERSION_URLS[current_version]

        except KeyError:
            msg = f"Link to data missing for current version: {current_version}."
            msg += f" Using default version: {InstallPackageData.DEFAULT_VERSION}"

            data_link = InstallPackageData.DEFAULT_VERSION

            print(msg)

        # retrieve content from URL
        print(
            f"Downloading example data for stitches version {current_version}.  This may take a few minutes..."
        )
        response = requests.get(data_link)

        with zipfile.ZipFile(BytesIO(response.content)) as zipped:
            # extract each file in the zipped dir to the project
            for f in zipped.namelist():
                extension = os.path.splitext(f)[-1]

                # Extract only the csv and nc files
                if all([len(extension) > 0, extension in (".csv", ".nc")]):
                    basename = os.path.basename(f)

                    # Check to see if tas-data is in the file path
                    if "tas-data" in f:
                        basename = os.path.join("tas-data", basename)

                    out_file = os.path.join(data_directory, basename)

                    # extract to a temporary directory to be able to only keep the file out of the dir structure
                    with tempfile.TemporaryDirectory() as tdir:
                        # extract file to temporary directory
                        zipped.extract(f, tdir)

                        # construct temporary file full path with name
                        tfile = os.path.join(tdir, f)

                        print(f"Unzipped: {out_file}")
                        # transfer only the file sans the parent directory to the data package
                        shutil.copy(tfile, out_file)


def install_package_data(data_dir: str = None):
    """
    Download and unpack Zenodo-minted stitches package data.

    This function matches the current installed stitches distribution and unpacks
    the data into the specified directory or the default data directory of the package.

    :param data_dir: Optional. Full path to the directory to store the data.
                     Default is the data directory of the package.
    :type data_dir: str

    :return: None
    """
    zen = InstallPackageData(data_dir=data_dir)

    zen.fetch_zenodo()



"""This module contains helper functions used throughout the stitches package."""

import os
from importlib import resources

import numpy as np
import pandas as pd


def combine_df(df1, df2):
    """
    Join two pandas data frames into a single data frame.

    :param df1: First pandas DataFrame.
    :param df2: Second pandas DataFrame.
    :return: A single pandas DataFrame resulting from the joining of df1 and df2.
    """
    incommon = df1.columns.intersection(df2.columns)
    if len(incommon) > 0:
        raise TypeError("a: df1 and df2 must have unique column names.")

    # Combine the two data frames with one another.
    df1["j"] = 1
    df2["j"] = 1
    out = df1.merge(df2)
    out = out.drop(columns="j")

    return out


def list_files(d):
    """
    Return the absolute path for all files in a directory, excluding .DS_Store files.

    :param d: Name of the directory.
    :type d: str
    :returns: List of file paths.
    :rtype: list
    """
    files = os.listdir(d)
    ofiles = []
    for i in range(0, len(files)):
        f = files[i]
        if not (".DS_Store" in f):
            ofiles.append(os.path.join(d, f))
    return ofiles


def selstr(a, start, stop):
    """
    Select elements of a string from start to stop index.

    :param a: Array containing a string.
    :type a: str
    :param start: First character index to select.
    :type start: int
    :param stop: Last character index to select.
    :type stop: int
    :returns: Array of strings.
    :rtype: list
    """
    if type(a) not in [str]:
        raise TypeError("a: must be a single string")

    out = []
    for i in range(start, stop):
        out.append(a[i])
    out = "".join(out)
    return out


def check_columns(data, names):
    """
    Check if a DataFrame contains all required columns.

    :param data: DataFrame to check.
    :type data: pd.DataFrame
    :param names: Set of required column names.
    :type names: set
    :raises TypeError: If `names` is not a set or if required columns are missing.
    """
    col_names = set(data.columns)
    if not (type(names) == set):
        raise TypeError("names must be a set.")

    if not (names.issubset(col_names)):
        raise TypeError(f'Missing columns from "{data}".')


def nrow(df):
    """
    Return the number of rows in the data frame.

    :param df: DataFrame to count rows for.
    :type df: pd.DataFrame
    :return: Number of rows in the data frame.
    :rtype: int
    """
    return df.shape[0]


def remove_obs_from_match(md, rm):
    """
    Return an updated matched data frame to prevent envelope collapse.

    This function is useful for preventing envelope collapse between
    generated and target ensembles by removing observations from the match.

    :param md: Matched data as a pandas DataFrame.
    :param rm: Data to remove as a pandas DataFrame.
    :return: Updated matched data frame as a pandas DataFrame.
    """
    rm = rm[
        [
            "target_year",
            "target_start_yr",
            "target_end_yr",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
        ]
    ].copy()
    rm["remove"] = True
    mergedTable = md.merge(rm, how="left")
    key = mergedTable["remove"].isnull()
    out = mergedTable.loc[key][
        [
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
        ]
    ]
    return out


def anti_join(x, y, bycols):
    """
    Return a DataFrame of the rows in `x` that do not appear in `y`.

    This function maintains only the columns of `x` with their original names,
    potentially in a different order. It performs an anti-join operation based
    on the specified columns.

    :param x: DataFrame to be filtered.
    :param y: DataFrame to filter against.
    :param bycols: Columns to perform the anti-join on.
    :return: A DataFrame containing the filtered result.
    """
    # Check the inputs
    check_columns(x, set(bycols))
    check_columns(y, set(bycols))

    # select only the entries of x that are not (['_merge'] == 'left_only') in y
    left_joined = x.merge(y, how="left", on=bycols, indicator=True).copy()
    left_only = (
        left_joined.loc[left_joined["_merge"] == "left_only"]
        .drop("_merge", axis=1)
        .copy()
    )

    # left_only has all the columns of x and y, with _x, _y appended to any that
    # had the same names. Want to return left_only with only the columns of x, but
    # which is bycols + anything _x (I think?)
    #
    # first, identify columns that end in _x, subset left_only to just those, and
    # rename the columns to remove the _x:
    _x_ending_cols = [col for col in left_only if col.endswith("_x")]
    left_only_x_ending_cols = left_only[_x_ending_cols].copy()
    new_names = list(
        map(lambda z: z.replace("_x", ""), left_only_x_ending_cols.columns)
    )
    left_only_x_ending_cols.columns = new_names
    #
    # concatenate those with the bycols:
    out = pd.concat([left_only[bycols], left_only_x_ending_cols], axis=1)

    # re-order the columns of out so that they are in the same order
    # as the columns of x
    cols_of_x_in_order = x.columns.copy()
    out = out[cols_of_x_in_order].copy()
    return out


def load_data_files(subdir):
    """
    Read in a list of data frames from a specified subdirectory.

    :param subdir: Subdirectory from which to load data files.
    :type subdir: str
    :return: A single pandas DataFrame object containing concatenated data from all files.
    """
    # Make sure the sub directory exists.
    path = resources.files("stitches") / subdir
    if not os.path.isdir(path):
        raise TypeError("subdir does not exist")

    # Find all of the files.
    files_to_process = list_files(path)
    raw_data = []

    # Read in the data & concatenate into a single data frame.
    for f in files_to_process:
        extension = os.path.splitext(f)[-1].casefold()

        if extension == ".csv":
            d = pd.read_csv(f)
        elif extension == ".pkl":
            d = pd.read_pickle(f)
        else:
            d = None
        raw_data.append(d)
    raw_data = pd.concat(raw_data)

    return raw_data



import unittest
from importlib import resources

import pandas as pd

from stitches.fx_match import (
    drop_hist_false_duplicates,
    internal_dist,
    match_neighborhood,
    shuffle_function,
)
from stitches.fx_util import nrow, remove_obs_from_match


class TestMatch(unittest.TestCase):
    def test_match_fxns(self):
        """
        Test the `match_neighborhood` function.

        This test ensures that the `match_neighborhood` function
        operates correctly.
        """
        # Read in some made up target data.
        path = resources.files("stitches") / "data" / "example" / "test-target_dat.csv"
        data = pd.read_csv(path)

        # Start by checking on a self test & without the drop hist constraint.
        self_match = match_neighborhood(data, data, tol=0)
        self.assertEqual(data.shape[0], self_match.shape[0])

        # Now try matching with the test archive
        path = resources.files("stitches") / "data" / "example" / "test-archive_dat.csv"
        archive = pd.read_csv(path)
        match1 = match_neighborhood(data, archive, tol=0)
        # We know this must be true because the test archive contains duplicates of the
        # test data, but when we increase the tolerance there should be even more matches
        # in the data frame.
        self.assertEqual(match1.shape[0], data.shape[0] * 2)
        match2 = match_neighborhood(data, archive, tol=0.1)
        self.assertTrue(match2.shape[0] > data.shape[0])

        "Test the remove_obs_from_match function"
        out1 = remove_obs_from_match(match1, match1.iloc[[2]])
        self.assertTrue(nrow(out1) < nrow(match1))

        "Test internal shuffle fxn"
        subset_data = data.head(10)
        out = shuffle_function(subset_data)
        self.assertTrue(subset_data.shape == out.shape)
        self.assertTrue(abs(subset_data.year - out.year).mean() > 0)

        "Test internal_dist"
        # Test against self, it should return exactly 1 entry & the distance values calculated
        # should be equal to 0, since calculating the distance against itself.
        dist1 = internal_dist(subset_data.fx[0], subset_data.dx[0], subset_data)
        self.assertTrue(nrow(dist1) == 1)
        self.assertTrue(dist1.dist_dx[0] == 0)
        self.assertTrue(dist1.dist_fx[0] == 0)
        self.assertTrue(dist1.dist_l2[0] == 0)

        "Test drop_hist_false_duplicates"
        # Read in the match test data.
        path = resources.files("stitches") / "data" / "example" / "test-match_w_dup.csv"
        match_data = pd.read_csv(path)
        cleaned = drop_hist_false_duplicates(match_data)
        self.assertTrue(nrow(match_data) > nrow(cleaned))
        hist_target_experiments = cleaned[cleaned["target_start_yr"] <= 2020][
            "target_experiment"
        ].unique()
        self.assertEqual(len(hist_target_experiments), 1)


if __name__ == "__main__":
    unittest.main()



"""Module for generating package data for the stitches library.

This module contains functions to generate the tas archive, matching archive,
and the table of pangeo files. It is used to create the CSV files included in
the prebuilt stitches package.
"""

import stitches.make_matching_archive as mk_match
import stitches.make_pangeo_table as mk_pangeo
import stitches.make_tas_archive as mk_tas


def generate_pkg_data(
    smoothing_window=9,
    chunk_window=9,
    add_staggered=False,
    anomaly_startYr=1995,
    anomaly_endYr=2014,
):
    """
    Generate all internal package data for stitches.

    This function creates the tas archive, matching archive, and the table of
    pangeo files. It generates all of the CSV files included in the prebuilt
    stitches package and may produce temporary files during the process.

    :param smoothing_window: The smoothing window size.
    :type smoothing_window: int
    :param chunk_window: The chunk window size.
    :type chunk_window: int
    :param add_staggered: Flag to add staggered output.
    :type add_staggered: bool
    :param anomaly_startYr: The start year for anomaly calculation.
    :type anomaly_startYr: int
    :param anomaly_endYr: The end year for anomaly calculation.
    :type anomaly_endYr: int
    :return: None
    """
    # This takes several hours to run.
    mk_tas(anomaly_startYr=anomaly_startYr, anomaly_endYr=anomaly_endYr)

    # These two functions run quickly.
    mk_match(
        smoothing_window=smoothing_window,
        chunk_window=chunk_window,
        add_staggered=add_staggered,
    )
    mk_pangeo()

    return None



"""Module for accessing package data within the stitches package."""

import os
from importlib import resources

import xarray as xr

__all__ = ["fetch_quickstarter_data"]


def fetch_quickstarter_data(variable: str) -> xr.Dataset:
    """
    Fetch a quickstarter NetCDF dataset as an xarray object.

    :param variable: Target variable name.
    :type variable: str

    :return: Xarray Dataset for example data.
    :rtype: xr.Dataset
    """
    variable_lower = variable.casefold()

    if variable_lower not in ("tas", "pr"):
        raise KeyError(
            f"Variable '{variable}' not a valid option.  Choose from 'tas' or 'pr'."
        )

    f = (
        resources.files("stitches")
        / "data"
        / f"stitched_CanESM5_{variable_lower}_ssp245~r1i1p1f1~1.nc"
    )

    return xr.open_dataset(f)



import os
import unittest

import numpy as np
import pandas as pd

from stitches.fx_util import (
    anti_join,
    check_columns,
    combine_df,
    list_files,
    load_data_files,
    nrow,
    selstr,
)


class TestUtil(unittest.TestCase):
    """Unit tests for utility functions in the `stitches` package."""

    TABLE_A = pd.DataFrame(data={"col1": [1, 2, 3, 4], "col2": [3, 4, 5, 6]})
    TABLE_B = pd.DataFrame(data={"col1": [3, 4, 9], "col2": [5, 6, 10]})
    TABLE_C = pd.DataFrame(data={"col1": ["a", "b"], "col2": [4, 4]})
    TABLE_D = pd.DataFrame(data={"col3": ["a", "b", "c"], "col4": [3, 4, 5]})

    def test_check_columns(self):
        """Test the check_columns function for proper type checking."""
        self.assertTrue(
            isinstance(check_columns(self.TABLE_A, set(self.TABLE_A)), type(None))
        )
        with self.assertRaises(TypeError):
            check_columns(self.TABLE_A, self.TABLE_A.columns)
        with self.assertRaises(TypeError):
            check_columns(self.TABLE_A, {"fake"})

    def test_nrow(self):
        """Test the `nrow` function to ensure it returns the correct number of rows."""
        self.assertTrue(nrow(self.TABLE_A), 4)
        doubble_A = pd.concat([self.TABLE_A, self.TABLE_A])
        self.assertTrue(nrow(doubble_A), nrow(self.TABLE_A) * 2)

    def test_selstr(self):
        """Test the `selstr` function for substring extraction."""
        self.assertEqual(selstr("abcd", 0, 2), "ab")
        self.assertEqual(selstr("abcd", 0, 1), "a")
        with self.assertRaises(TypeError):
            selstr(123, 1, 2)
        with self.assertRaises(TypeError):
            selstr({"abcd", "abcd"}, 1, 2)

    def test_anti_join(self):
        """
        Test the `anti_join` function to ensure it returns the correct DataFrame.

        This test verifies that the `anti_join` function correctly returns a DataFrame
        that contains only the rows from the first input DataFrame that do not have
        matching key values in the second input DataFrame.
        """
        tableA = pd.DataFrame(
            np.random.rand(4, 3), pd.Index(list("abcd"), name="Key"), ["A", "B", "C"]
        ).reset_index()
        tableB = pd.DataFrame(
            np.random.rand(4, 3), pd.Index(list("aecf"), name="Key"), ["A", "B", "C"]
        ).reset_index()
        tableC = anti_join(tableA, tableB, bycols=["Key"])
        self.assertEqual(tableC.shape, (2, 4), "unexpected shape returned by anti_join")

        tableD = anti_join(tableB, tableA, bycols=["Key"])
        self.assertEqual(tableD.shape, (2, 4), "unexpected shape returned by anti_join")

    def test_combine_df(self):
        """Test the `combine_df` function for DataFrame combination."""
        TABLE_E = combine_df(self.TABLE_C, self.TABLE_D)
        self.assertEqual(nrow(TABLE_E), nrow(self.TABLE_C) * nrow(self.TABLE_D))
        with self.assertRaises(TypeError):
            combine_df(self.TABLE_C, self.TABLE_C)

    def test_file_fxns(self):
        """
        Test the `list_files` and `load_files` functions.
        """

        # Make sure that all of the files returned by the list_files function are all true.
        test_dir = "."
        out = list_files(test_dir)
        self.assertTrue(all(list(map(os.path.exists, out))))

        # test failure
        with self.assertRaises(TypeError):
            load_data_files("fake")


if __name__ == "__main__":
    unittest.main()



"""Module for processing weighted global mean temperature from Pangeo CMIP6 results."""

import os
from importlib import resources

import pandas as pd
from tqdm import tqdm

import stitches.fx_data as data
import stitches.fx_pangeo as pangeo
import stitches.fx_util as util


def join_exclude(dat, drop):
    """Drop some rows from a data frame.

    :param dat:   pd data frame containing the data that needs to be dropped.
    :param drop: pd data frame containing the data to drop.
    :return:    pd data frame that is subset.
    """
    dat = dat.copy()
    drop = drop.copy()

    # Get the column names that two data frames have
    # in common with one another.
    in_common = list(
        set(dat.columns) & set(drop.columns)
    )  # figure out what columns are in common between the two dfs
    drop[
        "drop"
    ] = 1  # add an indicator column to indicate which rows need to be dropped
    out = dat.merge(drop, how="left", on=in_common)

    out = out.loc[out["drop"].isna()]  # remove the entries that need to be dropped
    out = out[dat.columns]  # select the columns

    return out


def rbind(dat1, dat2):
    """Combine two data frames together even if a data frame is empty.

    :param dat1:    data frame of values
    :type dat1:     pandas.core.frame.DataFrame

    :param dat2:    data frame of values
    :type dat2:     pandas.core.frame.DataFrame

    :return: a singular data frame
    """
    if util.nrow(dat1) == 0:
        return dat2

    if util.nrow(dat2) == 0:
        return dat1

    out = pd.concat([dat1, dat2])
    return out


def get_global_tas(path):
    """
    Calculate the weighted annual global mean temperature.

    This function computes the weighted annual global mean surface air temperature
    from CMIP6 files stored on Pangeo.

    :param path: A Zarr store path to the CMIP6 files.
    :type path: str
    :return: Path to the file containing the weighted global mean temperature.
    """
    temp_dir = resources.files("stitches") / "data" / "temp-data"

    if not os.path.isdir(temp_dir):
        os.mkdir(temp_dir)

    tag = path.replace("/", "_")
    file_name = tag.replace("gs:__", "") + "temp.csv"
    ofile = temp_dir + "/" + file_name

    if not os.path.isfile(ofile):
        # Download the CMIP data & calculate the weighted annual global mean .
        d = pangeo.fetch_nc(path)
        global_mean = data.global_mean(d)
        annual_mean = global_mean.coarsen(time=12).mean()

        # Format the CMIP meta data & global means, then combine into a single data frame.
        meta = data.get_ds_meta(d)
        t = annual_mean["time"].dt.strftime("%Y%m%d").values
        year = list(map(lambda x: util.selstr(x, start=0, stop=4), t))

        val = annual_mean["tas"].values
        d = {"year": year, "value": val}
        df = pd.DataFrame(data=d)
        out = util.combine_df(meta, df)

        # Write the output
        out.to_csv(ofile, index=False)
        return ofile
    else:
        return ofile


def calculate_anomaly(data, startYr=1995, endYr=2014):
    """
    Convert CMIP absolute temperature data to anomalies relative to a reference period.

    This function transforms a DataFrame containing CMIP absolute temperature
    data into a DataFrame of temperature anomalies. Anomalies are calculated
    relative to a time-averaged value from a specified reference period.

    :param data: A DataFrame of the CMIP absolute temperature.
    :type data: pandas.core.frame.DataFrame
    :param startYr: The first year of the reference period. Defaults to 1995.
    :type startYr: int
    :param endYr: The final year of the reference period. Defaults to 2014.
    :type endYr: int
    :return: A DataFrame of CMIP temperature anomalies relative to the reference period.
    """
    # Inputs
    util.check_columns(
        data, {"variable", "experiment", "ensemble", "model", "year", "value"}
    )
    to_use = data[["model", "experiment", "ensemble", "year", "value"]].copy()

    # Calculate the average value for the reference period defined
    # by the startYr and ednYr arguments.
    # The default reference period is set from 1995 - 2014.
    #
    # Start by subsetting the data so that it only includes values from the specified reference period.
    to_use["year"] = to_use["year"].astype(int)
    to_use = to_use[to_use["experiment"] == "historical"]
    subset_data = to_use[to_use["year"].between(startYr, endYr)]

    # Calculate the time-averaged reference value for each ensemble
    # realization. This reference value will be used to convert from absolute to
    # relative temperature.
    reference_values = (
        subset_data.groupby(["model", "ensemble"])
        .agg({"value": lambda x: sum(x) / len(x)})
        .reset_index()
        .copy()
    )
    reference_values = reference_values.rename(columns={"value": "ref_values"})

    # Combine the dfs that contain the absolute temperature values with the reference values.
    # Then calculate the relative temperature.
    merged_df = data.merge(reference_values, on=["model", "ensemble"], how="inner")
    merged_df["value"] = merged_df["value"] - merged_df["ref_values"]
    merged_df = merged_df.drop(columns="ref_values")

    return merged_df


def paste_historical_data(input_data):
    """
    Paste historical data into each future scenario.

    This function appends the appropriate historical data to each future scenario,
    ensuring that, for example, SSP585 realization 1 contains data from 1850-2100.

    :param input_data: A DataFrame of the CMIP absolute temperature.
    :type input_data: pandas.core.frame.DataFrame
    :return: A DataFrame of the smoothed time series with a rolling mean applied.
    """
    # Relabel the historical values so that there is a continuous rolling mean between the
    # historical and future values.
    # #######################################################
    # Create a subset of the non historical & future scns
    other_exps = ["1pctCO2", "abrupt-2xCO2", "abrupt-4xCO2"]
    other_data = input_data[input_data["experiment"].isin(other_exps)]

    # Subset the historical data
    historical_data = input_data[input_data["experiment"] == "historical"].copy()

    # Create a subset of the future data
    fut_exps = [
        "ssp126",
        "ssp245",
        "ssp370",
        "ssp585",
        "ssp534-over",
        "ssp119",
        "ssp434",
        "ssp460",
    ]
    future_data = input_data[input_data["experiment"].isin(fut_exps)]
    future_scns = set(future_data["experiment"].unique())

    frames = []
    for scn in future_scns:
        d = historical_data.copy()
        d["experiment"] = scn
        frames.append(d)

    frames.append(future_data)
    frames.append(other_data)
    data = pd.concat(frames)

    d = (
        data.groupby(["variable", "experiment", "ensemble", "model", "year"])["value"]
        .agg("mean")
        .reset_index()
    )

    return d


def make_tas_archive(anomaly_startYr=1995, anomaly_endYr=2014):
    """
    Create the archive from Pangeo-hosted CMIP6 data.

    This function processes CMIP6 data hosted on Pangeo to create an archive of
    temperature anomaly files. It calculates anomalies based on a specified reference
    period.

    :param anomaly_startYr: Start year of the reference period for anomaly calculation.
    :type anomaly_startYr: int
    :param anomaly_endYr: End year of the reference period for anomaly calculation.
    :type anomaly_endYr: int
    :return: List of paths to the created tas files.
    :rtype: list
    """
    # Get the pangeo table of contents.
    df = pangeo.fetch_pangeo_table()

    # Subset the monthly tas data these are the files that we will want to process
    # for the tas archive.
    xps = [
        "historical",
        "1pctCO2",
        "abrupt-4xCO2",
        "abrupt-2xCO2",
        "ssp370",
        "ssp245",
        "ssp119",
        "ssp434",
        "ssp460",
        "ssp126",
        "ssp585",
        "ssp534-over",
    ]
    df = df.loc[
        (df["experiment_id"].isin(xps))
        & (df["table_id"] == "Amon")  # experiments of interest
        & (df["grid_label"] == "gn")  # monthly data
        & (  # we are only interested in the results returned in the native
            df["variable_id"] == "tas"
        )
        & (df["member_id"].str.contains("p1"))  # select temperature data
    ]  # select only the members of the p1 physics group

    # For each of the CMIP6 files to calculate the global mean temperature and write the
    # results to the temporary directory.
    print("Downloading tas data from pangeo...")
    to_download = tqdm(df.zstore.values)
    files = list(map(get_global_tas, to_download))

    print("Processing data this may take a moment.")
    # Clean Up & Quality Control
    #
    # Find all of the files and read in the data, store as a single data frame.
    raw_data = pd.concat(list(map(pd.read_csv, files)))

    # Note that the first three steps only apply to the historical & ssp experiments,
    # the idealized experiments do not need to go through these steps.
    #
    # First round of cleaning check the historical dates.
    # Make sure that the historical run starts some time
    # before 1855 & that that it runs until 2014.
    # Subset the Hector historical
    his_info = (
        raw_data.loc[(raw_data["experiment"] == "historical")]
        .groupby(["model", "experiment", "ensemble"])["year"]
        .agg(["min", "max"])
        .reset_index()
    )
    his_info["min"] = his_info["min"].astype(int)
    his_info["max"] = his_info["max"].astype(int)

    # Make sure the start date is some times before 1855.
    start_yr = his_info[his_info["min"] > 1855].copy()
    to_remove = start_yr[["model", "experiment", "ensemble"]]

    # Make sure that all of historical have data up until 2014
    end_yr = his_info[his_info["max"] < 2014].copy()
    to_remove = rbind(to_remove, end_yr[["model", "experiment", "ensemble"]])
    clean_d1 = join_exclude(raw_data, to_remove)

    # Second round of cleaning check the future dates.
    # Make sure that the future scenarios start at 2015 & run beyond 2100.
    fut_exps = [
        "ssp245",
        "ssp126",
        "ssp585",
        "ssp119",
        "ssp370",
        "ssp434",
        "ssp534-over",
        "ssp460",
    ]
    fut_info = (
        clean_d1.loc[(clean_d1["experiment"].isin(fut_exps))]
        .groupby(["model", "experiment", "ensemble"])["year"]
        .agg(["min", "max"])
        .reset_index()
    )
    fut_info["min"] = fut_info["min"].astype(int)
    fut_info["max"] = fut_info["max"].astype(int)

    # If the future scenario starts after 2015 drop it.
    start_yr = fut_info[fut_info["min"] > 2015].copy()
    to_remove = start_yr[["model", "experiment", "ensemble"]]

    # Make sure the future scenario runs until 2098 otherwise drop it.
    end_yr = fut_info[fut_info["max"] < 2098].copy()
    to_remove = rbind(to_remove, end_yr[["model", "experiment", "ensemble"]])
    clean_d2 = join_exclude(clean_d1, to_remove)

    # Third round of clean up
    # Make sure that there is data from the historical experiment for each ensemble member with
    # future results.
    exp_en_mod = clean_d2[["experiment", "ensemble", "model"]].drop_duplicates().copy()

    # Separate the data frame of the experiment / ensemble / model information into
    # the historical and non historical experiments.
    hist_ensemble = (
        exp_en_mod.loc[exp_en_mod["experiment"] == "historical"][["model", "ensemble"]]
        .drop_duplicates()
        .copy()
    )
    non_hist_ensemble = (
        exp_en_mod[exp_en_mod["experiment"] != "historical"][["model", "ensemble"]]
        .drop_duplicates()
        .copy()
    )
    # use an inner join to select the historical ensemble that have future results as well as
    # the future results have have historical results.
    to_keep = non_hist_ensemble.merge(
        hist_ensemble, how="inner", on=["ensemble", "model"]
    )

    # Update the raw data table to only include the model / ensembles members that have both a
    # historical and non historical ensemble realization.
    clean_d3 = clean_d2.merge(to_keep, how="inner")

    # Before the fourth round of clean up add back in the idealized experiment results.
    idealized_exps = {"1pctCO2", "abrupt-2xCO2", "abrupt-4xCO2"}
    idealized_dat = raw_data.loc[raw_data["experiment"].isin(idealized_exps)]
    clean_d3 = rbind(clean_d3, idealized_dat)

    # Fourth round of cleaning make sure there are no missing dates.
    yrs = (
        clean_d2.groupby(["model", "experiment", "ensemble"])["year"]
        .agg(["min", "max", "count"])
        .reset_index()
    )
    yrs["min"] = yrs["min"].astype(int)
    yrs["max"] = yrs["max"].astype(int)
    yrs["count"] = yrs["count"].astype(int)
    yrs["diff"] = (yrs["max"] - yrs["min"]) + 1
    yrs["diff"] = yrs["diff"].astype(int)
    to_remove = yrs[yrs["diff"] != yrs["count"]]
    clean_d4 = join_exclude(clean_d3, to_remove).copy()

    # Order the data frame to make sure that all of the years are in order.
    cleaned_data = (
        clean_d4.sort_values(
            by=["variable", "experiment", "ensemble", "model", "year"]
        ).reset_index(drop=True)
    ).copy()

    # Format Data
    #
    # In this section convert from absolute value to an anomaly & concatenate the historical data
    # with the future scenarios.
    data_anomaly = calculate_anomaly(
        cleaned_data, startYr=anomaly_startYr, endYr=anomaly_endYr
    ).copy()
    data = paste_historical_data(data_anomaly)
    data = data.sort_values(by=["variable", "experiment", "ensemble", "model", "year"])
    data = data[
        ["variable", "experiment", "ensemble", "model", "year", "value"]
    ].reset_index(drop=True)

    # Add the z store values to the data frame.
    # Get the pangeo table of contents & assert that the data frame exists and contains information.
    df = pangeo.fetch_pangeo_table()
    df = df.loc[
        (df["table_id"] == "Amon")
        & (df["grid_label"] == "gn")  # monthly data
        & (  # we are only interested in the results returned in the native
            df["variable_id"] == "tas"
        )
        & (df["member_id"].str.contains("p1"))  # select temperature data
    ].copy()  # select only the members of the p1 physics group

    if len(df) <= 0:
        raise Exception("Unable to connect to pangeo, make sure to disconnect from VP")

    # Format the pangeo data frame so that it reflects the contents of data.
    pangeo_df = df[
        ["source_id", "experiment_id", "member_id", "variable_id", "zstore"]
    ].drop_duplicates()
    pangeo_df = pangeo_df.rename(
        columns={
            "source_id": "model",
            "experiment_id": "experiment",
            "member_id": "ensemble",
            "variable_id": "variable",
        }
    )

    # Add the zstore file information to the data frame via a left join.
    data = data.merge(
        pangeo_df, on=["variable", "experiment", "ensemble", "model"], how="inner"
    )

    # Modify the zstore path names to replace the future scn string with historical.
    # TODO replace this for loop it is pretty slow
    new_zstore = []
    for i in data.index:
        # Select the row from the data frame.
        row = data.loc[i]

        # Check to see if the zstore needs to be changed based on if it is a future experiment.
        fut_exps = {
            "ssp119",
            "ssp126",
            "ssp245",
            "ssp370",
            "ssp434",
            "ssp460",
            "ssp534-over",
            "ssp585",
        }
        change = row["experiment"] in fut_exps
        if change:
            new = row["zstore"].replace(row["experiment"], "historical")
        else:
            new = row["zstore"]

        new_zstore.append(new)

    data["zstore"] = new_zstore

    # Save a copy of the tas values, these are the value that will be used to get the
    # tas data chunks. Note that this file has to be compressed so will need to read in
    # using pickle_utils.load()

    files = []
    tas_data_dir = resources.files("stitches") / "data" / "tas-data"
    for name, group in data.groupby(["model"]):
        path = tas_data_dir + "/" + name + "_tas.csv"
        files.append(path)
        group.to_csv(path, index=False)

    print("Global tas data complete")

    return files



"""Module for creating a Pangeo table for use in the stitches package."""

# Save a copy of the pangeo table, this has information about the z store file.
from importlib import resources

import pandas as pd

import stitches.fx_pangeo as pangeo


def make_pangeo_table():
    """
    Create a copy of the Pangeo files that have corresponding entries in the matching archive.

    This function is used in the stitching process to ensure that only relevant Pangeo files
    are considered. It writes out a file to the package data directory.

    :return: None
    """
    # Using the information about what experiment/ensemble/models that are available for matching.
    archive_path = resources.files("stitches") / "data" / "matching_archive.csv"
    tas_exp_model = (
        pd.read_csv(archive_path)[["experiment", "ensemble", "model"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )

    # For the future experiments make sure we add in the historical experiment info.
    fut_exps = ["ssp126", "ssp245", "ssp370", "ssp585", "ssp534-over"]
    hist_info = tas_exp_model.loc[tas_exp_model["experiment"].isin(fut_exps)].copy()[
        ["model", "ensemble"]
    ]
    hist_info["experiment"] = "historical"
    hist_info = hist_info.drop_duplicates()
    to_keep = pd.concat([tas_exp_model, hist_info], axis=0)
    to_keep = to_keep.reset_index(drop=True)

    # Get the complete data table from pangeo
    dat = pangeo.fetch_pangeo_table()
    pangeo_table = dat.loc[dat["grid_label"] == "gn"][
        ["source_id", "experiment_id", "member_id", "variable_id", "zstore", "table_id"]
    ].copy()
    pangeo_table = pangeo_table.rename(
        columns={
            "source_id": "model",
            "experiment_id": "experiment",
            "member_id": "ensemble",
            "variable_id": "variable",
            "zstore": "zstore",
            "table_id": "domain",
        }
    )
    pangeo_table = pangeo_table.reset_index(drop=True)

    # Subset the pangeo table so that it only contains information for what we have deemed as good data.
    final_pangeo_table = pangeo_table.merge(
        to_keep, how="inner", on=["model", "experiment", "ensemble"]
    )

    # Write the file out
    out_file = resources.files("stitches") / "data" / "pangeo_table.csv"
    final_pangeo_table.to_csv(out_file, index=False)

    # Return
    return None


def make_pangeo_comparison():
    """
    Create a copy of the entire Pangeo archive for testing.

    This function is used to check for updates in the Pangeo archive. If an update is
    detected, it may suggest updating the internal package data.

    :return: None. Writes a file to package data.
    """
    dat = pangeo.fetch_pangeo_table()
    out_file = resources.files("stitches") / "data" / "pangeo_comparison_table.csv"
    dat.to_csv(out_file, index=False)
    # Return
    return None



import unittest

import pandas as pd
import xarray as xr

from stitches.fx_pangeo import fetch_nc, fetch_pangeo_table


class TestPangeo(unittest.TestCase):
    """
    A test case for Pangeo-related functions.

    This test class is used to run tests for functions that interact with the Pangeo data archive.
    It includes a flag to run tests for continuous integration or for all cases.
    """

    RUN = "ci"

    def test_pangeo_fn(self):
        """Test Pangeo-related functions for continuous integration or full cases."""
        if TestPangeo.RUN == "ci":
            self.assertEqual(0, 0)
        else:
            # Get the table of all pangeo contents as a data frame
            ptable = fetch_pangeo_table()
            self.assertEqual(
                type(ptable),
                pd.core.frame.DataFrame,
                "fetch_pangeo_table did not return a data frame",
            )

            # Select a single files from the pangeo cataog to import
            import_this = ptable.loc[
                (ptable["table_id"] == "Amon")
                & (ptable["activity_id"] == "ScenarioMIP")
            ].reset_index(drop=True)
            path = import_this["zstore"][0]
            out = fetch_nc(path)
            self.assertEqual(
                type(out), xr.core.dataset.Dataset, "problem with fetch_nc"
            )


if __name__ == "__main__":
    unittest.main()



"""
The setup.py file for the stitches package.

The stitches package provides tools for stitching together climate model output
into a single, coherent dataset.
"""


import re

from setuptools import find_packages, setup


def requirements():
    """
    Read and parse the 'requirements.txt' file.

    Returns a list of requirements specified in the 'requirements.txt' file.

    :return: A list of package requirements.
    :rtype: list
    """
    with open("requirements.txt") as f:
        return f.read().split()


version = re.search(
    r"__version__ = ['\"]([^'\"]*)['\"]", open("stitches/_version.py").read(), re.M
).group(1)

setup(
    name="stitches-emulator",
    version=version,
    packages=find_packages(),
    url="https://github.com/JGCRI/stitches",
    license="BSD 2-Clause",
    author="Abigail Snyder, Kalyn Dorheim, Claudia Tebaldi",
    author_email="abigail.snyder@pnnl.gov, kalyn.dorheim@pnnl.gov, claudia.tebaldi@pnnl.gov",
    description="Amalgamate existing climate data to create monthly climate variable fields",
    python_requires=">=3.9.0",
    include_package_data=True,
    install_requires=requirements(),
    extras_require={
        "dev": [
            "pytest>=7.4.3",
            "build>=0.5.1",
            "nbsphinx>=0.8.6",
            "setuptools>=57.0.0",
            "sphinx>=7",
            "sphinx_book_theme>=1",
            "sphinx-click>=5.1",
            "sphinx_copybutton>=0.5",
            "twine~=3.4.1",
            "pre-commit>=3.6.0",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
)



"""
The `fx_match` module is responsible for matching target climate model outputs with archived model outputs.

It includes functions to calculate distances between target and archive data points and to
select the best matches based on those distances.
"""

import numpy as np
import pandas as pd

import stitches.fx_util as util


# Internal fx
def internal_dist(fx_pt, dx_pt, archivedata, tol=0):
    """
    Calculate the Euclidean distance between target and archive values.

    This function calculates the Euclidean distance between the target values (fx and dx)
    and the archive values contained in the dataframe. It is used to select which
    archive values best match the target values. To ensure consistent units across
    all dimensions, dx is updated to be windowsize*dx with units of degC, resulting
    in a distance metric (Euclidean/l2) in units of degC. The choice of unit consistency
    could be made optional via an argument and if-statement.

    :param fx_pt: A single value of the target fx value.
    :param dx_pt: A single value of the target dx value.
    :param archivedata: A dataframe of the archive fx and dx values.
    :param tol: A tolerance for the neighborhood of matching; defaults to 0 degC,
                returning only the nearest neighbor.
    :return: A dataframe with the target data and the corresponding matched archive data.
    """
    # Check the inputs
    util.check_columns(
        archivedata,
        {
            "model",
            "experiment",
            "variable",
            "ensemble",
            "start_yr",
            "end_yr",
            "year",
            "fx",
            "dx",
        },
    )

    # Compute the window size of the archive data to use to update
    # dx values to be windowsize*dx so that it has units of degC
    windowsize = max(archivedata["end_yr"] - archivedata["start_yr"])

    # Update the names of the columns in the archive data.
    dist = archivedata.copy()
    dist.columns = "archive_" + dist.columns

    # calculate the euclidean distance between the target point
    # and the archive observations.
    # Calculating distance d(target, archive_i) for each point i in the archive.
    # calculating the distance in the dx and fx dimensions separately because
    # now we want to track those in addition to the l2 distance.
    dist["dist_dx"] = windowsize * abs(dist["archive_dx"] - dx_pt)
    dist["dist_fx"] = abs(dist["archive_fx"] - fx_pt)
    dist["dist_l2"] = (dist["dist_fx"] ** 2 + dist["dist_dx"] ** 2) ** 0.5

    # this returns the first minimum run into, which is not how we are going to want to do it,
    # we will want some way to keep track of the min and combine to have different realizations
    # or have some random generation. But for now I think that this is probably sufficient.
    min_dist = dist["dist_l2"][np.argmin(dist["dist_l2"])]
    dist_radius = min_dist + tol
    index = np.where(dist["dist_l2"].values <= dist_radius)
    out = dist.loc[index]

    out["target_fx"] = fx_pt
    out["target_dx"] = dx_pt

    return out


# Internal fx
def shuffle_function(dt):
    """
    Randomly shuffle the deck to assist with the matching process.

    :param dt: A DataFrame of archive values used in the matching process.
    :return: A DataFrame with rows in random order.
    """
    nrow = dt.shape[0]
    out = dt.sample(nrow, replace=False)
    out = out.reset_index(drop=True)
    return out


# Internal fx
def drop_hist_false_duplicates(matched_data):
    """
    Remove false duplicate matches in the historical period.

    This function is used to remove false duplicate matches in the historical period.
    For example, if the target year 1850 gets data from 1872 from realization 13 of
    SSP126 and SSP585, the metadata of these archive values are different, but the
    actual data values are identical because the same historical data was pasted into
    every experiment. This function keeps only the first match.

    :param matched_data: pandas DataFrame returned from match_neighborhood.
    :return: DataFrame with the same structure as the input, with false duplicates
             in the historical period dropped.
    """
    # Subset the idealized runs, since these are not concatenated with the historical time series
    # they can be left alone.
    idealized_exps = ["1pctCO2", "abrupt-4xCO2", "abrupt-2xCO2"]
    idealized_matched_data = matched_data.loc[
        matched_data["archive_experiment"].isin(idealized_exps)
    ].copy()

    # Now select the non idealized runs, these are the time series that are a combination of the
    # historical and future time series.
    fut_matched_data = matched_data.loc[
        ~matched_data["archive_experiment"].isin(idealized_exps)
    ].copy()

    # Determine the historical cut off year based on the size of the chunks.
    cut_off_yr = (
        2015
        - max(fut_matched_data["target_end_yr"] - fut_matched_data["target_start_yr"])
        / 2
    )

    # Because smoothing with window =9 has occurred,
    # historical is actually 2010 or earlier: the chunks
    # that had purely historical data in them and none
    # from the future when smoothing.
    subset_df = fut_matched_data.loc[
        fut_matched_data["target_year"] <= cut_off_yr
    ].copy()

    # Only operate if there are any historic years to deal with:
    if util.nrow(subset_df) > 0:
        # parse out information about the ssp experiment id
        subset_df["idvalue"] = list(
            map(
                lambda x: int(x.split("p")[1].replace("-over", "")),
                subset_df["archive_experiment"],
            )
        )
        # group the data frame by the target information, recall that for each target match
        # there might be more matches with the archive. Here we want to make sure that we do
        # not want to have multiple matches with experiments from the historical period
        # from the archive.
        grouped_dat = subset_df.groupby(
            [
                "target_variable",
                "target_experiment",
                "target_ensemble",
                "target_model",
                "target_start_yr",
                "target_end_yr",
                "target_year",
                "target_fx",
                "target_dx",
                "archive_ensemble",
                "archive_year",
            ]
        )
        dat = []
        for name, group in grouped_dat:
            min_id_value = min(group["idvalue"])
            dat.append(group.loc[group["idvalue"] == min_id_value])
        historical = pd.concat(dat)
        cols_to_keep = [
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
        ]
        historical = historical[cols_to_keep]

        # select the future matched data
        # TODO should it be 2010 or should it be the cut off year?
        matched = pd.concat(
            [
                fut_matched_data.loc[fut_matched_data["target_year"] > 2010].copy(),
                historical,
                idealized_matched_data,
            ]
        )
        matched = matched.sort_values("target_year")
        matched = matched.reset_index()
        cols_to_keep = [
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
            "archive_experiment",
            "archive_variable",
            "archive_model",
            "archive_ensemble",
            "archive_start_yr",
            "archive_end_yr",
            "archive_year",
            "archive_fx",
            "archive_dx",
            "dist_dx",
            "dist_fx",
            "dist_l2",
        ]
        matched = matched[cols_to_keep]
        return matched
    else:
        return matched_data


def match_neighborhood(
    target_data, archive_data, tol: float = 0, drop_hist_duplicates: bool = True
):
    """
    Calculate the Euclidean distance between target and archive data.

    This function takes data frames of target and archive data and calculates the
    Euclidean distance between the target values (fx and dx) and the archive values.

    :param target_data: Data frame of the target fx and dx values.
    :param archive_data: Data frame of the archive fx and dx values.
    :param tol: Tolerance for the neighborhood of matching. Defaults to 0 degC,
        meaning only the nearest-neighbor is returned. Must be a float.
    :param drop_hist_duplicates: Determines whether to consider historical values
        across SSP scenarios as duplicates (True) and drop all but one from matching,
        or to consider them as distinct points for matching (False). Defaults to True.
    :type drop_hist_duplicates: bool
    :return: Data frame with the target data and the corresponding matched archive data.
    """
    # Check the inputs of the functions
    if util.nrow(target_data) <= 0:
        raise TypeError("target_data is an empty data frame")
    if util.nrow(archive_data) <= 0:
        raise TypeError("archive_data is an empty data frame")
    util.check_columns(
        archive_data,
        {"experiment", "variable", "ensemble", "start_yr", "end_yr", "fx", "dx"},
    )
    util.check_columns(target_data, {"start_yr", "end_yr", "fx", "dx"})

    archive_data = archive_data.reset_index(drop=True).copy()

    # For every entry in the target data frame find its nearest neighbor from the archive data.
    # concatenate the results into a single data frame.
    rslt = map(
        lambda fx, dx: internal_dist(fx, dx, archivedata=archive_data, tol=tol),
        target_data["fx"],
        target_data["dx"],
    )
    matched = pd.concat(list(rslt))

    # Now add the information about the matches to the target data
    # Make sure it is clear which columns contain  data that comes from the target compared
    # to which ones correspond to the archive information.
    #
    # Make a copy of the target data to work with and use in creating the output matched data frame,
    # so that this function doesn't change the column names of one of its arguments.
    target_data_copied = target_data.copy()
    target_data_copied.columns = "target_" + target_data_copied.columns

    out = matched.merge(target_data_copied, how="left", on=["target_fx", "target_dx"])
    cols_to_keep = [
        "target_variable",
        "target_experiment",
        "target_ensemble",
        "target_model",
        "target_start_yr",
        "target_end_yr",
        "target_year",
        "target_fx",
        "target_dx",
        "archive_experiment",
        "archive_variable",
        "archive_model",
        "archive_ensemble",
        "archive_start_yr",
        "archive_end_yr",
        "archive_year",
        "archive_fx",
        "archive_dx",
        "dist_dx",
        "dist_fx",
        "dist_l2",
    ]
    out = out[cols_to_keep].drop_duplicates()

    if drop_hist_duplicates:
        out = drop_hist_false_duplicates(out)

    # if there are any nearest neighbor matches that are maybe still large,
    # warn the user that they will want to validate the outcome.

    # get the nearest neighbor match only for each target window
    grouped = out.groupby(
        [
            "target_variable",
            "target_experiment",
            "target_ensemble",
            "target_model",
            "target_start_yr",
            "target_end_yr",
            "target_year",
            "target_fx",
            "target_dx",
        ]
    )
    formatted_nn = pd.DataFrame()
    for name, group in grouped:
        df = group.copy()
        df = df[df["dist_l2"] == np.min(df["dist_l2"])].copy()
        formatted_nn = pd.concat([formatted_nn, df]).reset_index(drop=True)
        del df

    # subset to just the far away nearest neighbors
    formatted_nn = (
        formatted_nn[formatted_nn["dist_l2"] > 0.25].reset_index(drop=True).copy()
    )

    if not formatted_nn.empty:
        print("The following target windows have a nearest neighbor in T, dT space")
        print("that is more than 0.25degC away. This may or may not result in poor")
        print("matches and we recommend validation.")
        print(formatted_nn)
        print(
            "-----------------------------------------------------------------------------------------"
        )
    del formatted_nn
    del grouped

    return out



import unittest
from importlib import resources

import pandas as pd

from stitches.fx_match import match_neighborhood
from stitches.fx_recipe import (
    get_num_perms,
    permute_stitching_recipes,
    remove_duplicates,
)
from stitches.fx_util import check_columns


class TestRecipe(unittest.TestCase):
    """
    A collection of unittests for testing the recipe functionality within the
    stitches package. This includes tests for duplicate removal, permutation
    of stitching recipes, and other recipe-related functions.
    """

    TARGET_DATA = pd.DataFrame(
        data={
            "ensemble": ["r1i1p1f1"] * 28,
            "variable": ["tas"] * 28,
            "model": ["test_model"] * 28,
            "experiment": ["ssp245"] * 28,
            "start_yr": [
                1850,
                1859,
                1868,
                1877,
                1886,
                1895,
                1904,
                1913,
                1922,
                1931,
                1940,
                1949,
                1958,
                1967,
                1976,
                1985,
                1994,
                2003,
                2012,
                2021,
                2030,
                2039,
                2048,
                2057,
                2066,
                2075,
                2084,
                2093,
            ],
            "end_yr": [
                1858,
                1867,
                1876,
                1885,
                1894,
                1903,
                1912,
                1921,
                1930,
                1939,
                1948,
                1957,
                1966,
                1975,
                1984,
                1993,
                2002,
                2011,
                2020,
                2029,
                2038,
                2047,
                2056,
                2065,
                2074,
                2083,
                2092,
                2100,
            ],
            "year": [
                1854,
                1863,
                1872,
                1881,
                1890,
                1899,
                1908,
                1917,
                1926,
                1935,
                1944,
                1953,
                1962,
                1971,
                1980,
                1989,
                1998,
                2007,
                2016,
                2025,
                2034,
                2043,
                2052,
                2061,
                2070,
                2079,
                2088,
                2097,
            ],
            "fx": [
                -1.28986999,
                -1.26877615,
                -1.2150827,
                -1.19510465,
                -1.2504297,
                -1.12476316,
                -1.18010115,
                -1.19607191,
                -1.05674376,
                -1.06501778,
                -0.94348729,
                -0.95177229,
                -0.94919908,
                -0.95949291,
                -0.69856503,
                -0.56833508,
                -0.21722963,
                0.05139228,
                0.40485701,
                0.90463641,
                1.17029258,
                1.51889658,
                1.83794095,
                2.22997751,
                2.45954622,
                2.62914668,
                2.9379544,
                3.03012582,
            ],
            "dx": [
                -0.00336812,
                0.00931957,
                0.00475767,
                -0.01462664,
                0.01469504,
                -0.00829463,
                0.00480449,
                0.01124357,
                0.00170745,
                0.01226593,
                0.01214147,
                -0.01130722,
                -0.00533205,
                0.0263663,
                0.01746445,
                0.0110813,
                0.05199482,
                0.03731981,
                0.03515222,
                0.04676322,
                0.03548417,
                0.03356049,
                0.04786934,
                0.02252523,
                0.03451772,
                0.0222264,
                0.01888666,
                0.01404303,
            ],
        }
    )

    ARCHIVE_DATA = pd.DataFrame(
        data={
            "ensemble": (["r2i1p1f1"] * 28 + ["r3i1p1f1"] * 28),
            "variable": ["tas"] * 56,
            "model": ["test_model"] * 56,
            "experiment": ["ssp245"] * 56,
            "start_yr": [
                1850,
                1859,
                1868,
                1877,
                1886,
                1895,
                1904,
                1913,
                1922,
                1931,
                1940,
                1949,
                1958,
                1967,
                1976,
                1985,
                1994,
                2003,
                2012,
                2021,
                2030,
                2039,
                2048,
                2057,
                2066,
                2075,
                2084,
                2093,
            ]
            * 2,
            "end_yr": [
                1858,
                1867,
                1876,
                1885,
                1894,
                1903,
                1912,
                1921,
                1930,
                1939,
                1948,
                1957,
                1966,
                1975,
                1984,
                1993,
                2002,
                2011,
                2020,
                2029,
                2038,
                2047,
                2056,
                2065,
                2074,
                2083,
                2092,
                2100,
            ]
            * 2,
            "year": [
                1854,
                1863,
                1872,
                1881,
                1890,
                1899,
                1908,
                1917,
                1926,
                1935,
                1944,
                1953,
                1962,
                1971,
                1980,
                1989,
                1998,
                2007,
                2016,
                2025,
                2034,
                2043,
                2052,
                2061,
                2070,
                2079,
                2088,
                2097,
            ]
            * 2,
            "fx": [
                -1.09653809,
                -1.13315784,
                -1.0943591,
                -1.17177001,
                -1.32738625,
                -1.18827856,
                -1.19859187,
                -1.17866016,
                -1.01826316,
                -1.00823875,
                -0.86747065,
                -0.75741926,
                -0.95912482,
                -0.84078494,
                -0.70828715,
                -0.55041327,
                -0.28319994,
                0.09212478,
                0.60853009,
                1.00862082,
                1.3594721,
                1.78422761,
                2.01833223,
                2.34533018,
                2.65273186,
                2.96080091,
                2.96904659,
                3.0398682,
                -1.19284299,
                -1.23273612,
                -1.11309838,
                -1.20154639,
                -1.34276733,
                -1.28168792,
                -1.33553245,
                -1.20231452,
                -1.02608287,
                -0.94425834,
                -0.92477871,
                -0.8000345,
                -0.92952369,
                -0.97309169,
                -0.81131182,
                -0.61165323,
                -0.26776859,
                0.0471447,
                0.5355102,
                0.78498746,
                1.2937814,
                1.65184098,
                2.0555254,
                2.38665088,
                2.59874084,
                2.87927608,
                3.04260422,
                3.1910005,
            ],
            "dx": [
                0.01181598,
                -0.00976008,
                0.00192164,
                -0.0302527,
                0.01037644,
                0.00079727,
                0.01406466,
                0.00724873,
                0.00108201,
                0.02716604,
                -0.00360448,
                0.00981227,
                -0.03041208,
                0.02847502,
                0.01515118,
                0.01836129,
                0.05547686,
                0.03248089,
                0.06159063,
                0.03715717,
                0.04961695,
                0.0277128,
                0.03308734,
                0.04132718,
                0.02824231,
                0.01661322,
                0.00860536,
                -0.0009309,
                -0.00881434,
                0.0083732,
                0.00478882,
                -0.02305601,
                0.01004446,
                -0.00554876,
                0.00565863,
                0.01500687,
                0.01383536,
                0.01050035,
                0.00444087,
                0.00508502,
                -0.02231988,
                0.01092053,
                0.0227568,
                0.02147527,
                0.04722855,
                0.0478629,
                0.0360716,
                0.04215114,
                0.04714541,
                0.04151509,
                0.04368013,
                0.02965839,
                0.03212568,
                0.02365025,
                0.01311625,
                0.01300568,
            ],
        }
    )

    # Data relevant for remove_duplicates()
    # these are the duplicate matches that should result from match_neighborhood
    # with tol=0 on the above data sets.
    DUPLICATES = pd.DataFrame(
        data={
            "target_variable": ["tas"] * 4,
            "target_experiment": ["ssp245"] * 4,
            "target_ensemble": ["r1i1p1f1"] * 4,
            "target_model": ["test_model"] * 4,
            "target_start_yr": [1859, 1868, 2057, 2066],
            "target_end_yr": [1867, 1876, 2065, 2074],
            "target_year": [1863, 1872, 2061, 2070],
            "target_fx": [-1.26877615, -1.2150827, 2.22997751, 2.45954622],
            "target_dx": [0.00931957, 0.00475767, 0.02252523, 0.03451772],
            "archive_experiment": ["ssp245"] * 4,
            "archive_variable": ["tas"] * 4,
            "archive_model": ["test_model"] * 4,
            "archive_ensemble": ["r3i1p1f1", "r3i1p1f1", "r3i1p1f1", "r3i1p1f1"],
            "archive_start_yr": [1859, 1859, 2057, 2057],
            "archive_end_yr": [1867, 1867, 2065, 2065],
            "archive_year": [1863, 1863, 2061, 2061],
            "archive_fx": [-1.23273612, -1.23273612, 2.38665088, 2.38665088],
            "archive_dx": [0.0083732, 0.0083732, 0.02965839, 0.02965839],
            "dist_dx": [0.00757096, 0.02892424, 0.05706528, 0.03887464],
            "dist_fx": [0.03604003, 0.01765342, 0.15667337, 0.07289534],
            "dist_l2": [0.03682666, 0.03388591, 0.16674229, 0.08261337],
        }
    )

    # Target years r1-1863 and r1-1872 both match to the same archive point: r3-1863.
    # dist_l2 for r1-1872 = 0.034
    # dist_l2 for r1-1863 = 0.037
    # So r1-1872 keeps the match r3-1863 and r1-1863 needs to be rematched
    #
    # similarly targets r1-2061 and r1-2070 match to the same archive point: r3-2061.
    # dist_l2 for r1-2061 = 0.17
    # dist_l2 for r1-2070 = 0.08
    # So r1-2070 keeps the match and r1-2061 needs to be rematched.
    #
    # So we will explicitly check that remove_dupilcates() does this.
    REMATCHED_YEARS = [1863, 2061]

    NEW_MATCHES = pd.DataFrame(
        data={
            "target_variable": ["tas"] * 2,
            "target_experiment": ["ssp245"] * 2,
            "target_ensemble": ["r1i1p1f1"] * 2,
            "target_model": ["test_model"] * 2,
            "target_start_yr": [1859, 2057],
            "target_end_yr": [1867, 2065],
            "target_year": [1863, 2061],
            "target_fx": [-1.26877615, 2.22997751],
            "target_dx": [0.00931957, 0.02252523],
            "archive_experiment": ["ssp245"] * 2,
            "archive_variable": ["tas"] * 2,
            "archive_model": ["test_model"] * 2,
            "archive_ensemble": ["r2i1p1f1"] * 2,
            "archive_start_yr": [1886, 2057],
            "archive_end_yr": [1894, 2065],
            "archive_year": [1890, 2061],
            "archive_fx": [-1.32738625, 2.34533018],
            "archive_dx": [0.01037644, 0.04132718],
            "dist_dx": [0.00845496, 0.1504156],
            "dist_fx": [0.0586101, 0.11535267],
            "dist_l2": [0.05921681, 0.18955498],
        }
    )

    def test_get_num_perms(self):
        """
        Test the `get_num_perms` function to ensure it returns the correct number of permutations.

        This test verifies that the function returns a list of permutations for the match data,
        and that the list contains the expected columns.
        """
        # Read in the match test data.
        path = resources.files("stitches") / "data" / "example" / "test-match_w_dup.csv"
        match_data = pd.read_csv(path)

        # Test get_num_perms
        out = get_num_perms(match_data)
        self.assertEqual(type(out), list)
        # Throw an error if the output does not match what we would expect.
        check_columns(
            out[0],
            {
                "target_variable",
                "target_experiment",
                "target_ensemble",
                "target_model",
                "minNumMatches",
                "totalNumPerms",
            },
        )
        check_columns(
            out[1],
            {
                "target_variable",
                "target_experiment",
                "target_ensemble",
                "target_model",
                "target_start_yr",
                "target_end_yr",
                "target_year",
                "target_fx",
                "target_dx",
                "n_matches",
            },
        )
        self.assertEqual(len(out), 2, "Test get_num_perms")

    def test_remove_duplicates(self):
        """
        Test the remove_duplicates function for correct operation.

        Ensures that the function correctly identifies and removes duplicate
        entries from the dataset.
        """

        # Initial match data
        md = match_neighborhood(
            TestRecipe.TARGET_DATA, TestRecipe.ARCHIVE_DATA, tol=0.0
        )

        # # #####################################################
        # # Check to see if in the matched data frame if there are any repeated values.
        # md_archive = md[['archive_experiment', 'archive_variable', 'archive_model',
        #                            'archive_ensemble', 'archive_start_yr', 'archive_end_yr',
        #                            'archive_year', 'archive_fx', 'archive_dx']].copy()
        # duplicates = md.merge(md_archive[md_archive.duplicated()], how="inner").copy()
        #
        # # assert that the correct duplicate matches returned from match_neighborhood()
        # pd.testing.assert_frame_equal(duplicates, TestRecipe.DUPLICATES)
        #
        # # #####################################################

        cleaned_matched_data = remove_duplicates(md=md, archive=TestRecipe.ARCHIVE_DATA)

        # every target point but r1-1863 and r1-2061 should be identical to md
        pd.testing.assert_frame_equal(
            md[-md["target_year"].isin(TestRecipe.REMATCHED_YEARS)],
            cleaned_matched_data[
                -cleaned_matched_data["target_year"].isin(TestRecipe.REMATCHED_YEARS)
            ],
        )

        # and r1-1863 and r1-2061 should have been rematched with the resulting
        # matches described in TestRecipe.NEW_MATCHES
        pd.testing.assert_frame_equal(
            cleaned_matched_data[
                cleaned_matched_data["target_year"].isin(TestRecipe.REMATCHED_YEARS)
            ].reset_index(drop=True),
            TestRecipe.NEW_MATCHES,
        )

    def test_permute_stitching_recipes(self):
        """Test the permute_stitching_recipes function for correct operation."""
        # With tol < 0.17, the test data can only support one collapse free
        # recipe. So a message will be printed to that effect
        messy_rp1 = permute_stitching_recipes(
            N_matches=2,
            matched_data=match_neighborhood(
                TestRecipe.TARGET_DATA, TestRecipe.ARCHIVE_DATA, tol=0.07
            ),
            archive=TestRecipe.ARCHIVE_DATA,
            testing=True,
        )

        # Test that the recipe has the same number of time windows as the
        # target:
        self.assertEqual(
            len(messy_rp1),
            len(TestRecipe.TARGET_DATA),
            "drawn recipe does not have the same number of time windows as target ",
        )

        # Test that all of the recipe windows did in fact come from the archive:
        cols = [col for col in messy_rp1 if col.startswith("archive_")]
        archive_check = messy_rp1[cols]
        new_names = list(
            map(lambda x: x.replace("archive_", ""), archive_check.columns)
        )
        archive_check.columns = new_names
        self.assertTrue(
            len(archive_check.merge(TestRecipe.ARCHIVE_DATA)) == len(archive_check),
            "recipe has entries not from the archive somehow.",
        )

        # Test that no two target points got matched to the same archive point
        # (ie that remove_duplicates was called correctly in permute_stitching_recipes)
        md_archive = messy_rp1[
            [
                "archive_experiment",
                "archive_variable",
                "archive_model",
                "archive_ensemble",
                "archive_start_yr",
                "archive_end_yr",
                "archive_year",
                "archive_fx",
                "archive_dx",
            ]
        ]
        duplicates = messy_rp1.merge(md_archive[md_archive.duplicated()], how="inner")
        self.assertEqual(
            len(duplicates),
            0,
            "issue with call to remove_duplicates() in permute_stitching_recipe().",
        )

        # Test that our reproducible mode (testing=True) does produce the
        # same recipe:
        messy_rp2 = permute_stitching_recipes(
            N_matches=2,
            matched_data=match_neighborhood(
                TestRecipe.TARGET_DATA, TestRecipe.ARCHIVE_DATA, tol=0.07
            ),
            archive=TestRecipe.ARCHIVE_DATA,
            testing=True,
        )
        self.assertTrue(
            messy_rp2.equals(messy_rp1),
            "reproducible mode of permute_stitching_recipe failing 1",
        )

        # Test that turning it off produces something different:
        messy_rp3 = permute_stitching_recipes(
            N_matches=2,
            matched_data=match_neighborhood(
                TestRecipe.TARGET_DATA, TestRecipe.ARCHIVE_DATA, tol=0.07
            ),
            archive=TestRecipe.ARCHIVE_DATA,
            testing=False,
        )
        self.assertTrue(
            not messy_rp3.equals(messy_rp1),
            "reproducible mode of permute_stitching_recipe failing 2",
        )

        # Test that we don't have collapse when targeting multiple
        # ensemble members.
        # combine the TARGET and ARCHIVE data sets so we have more ensemble
        # members to work with:
        target2 = (
            pd.concat([TestRecipe.ARCHIVE_DATA, TestRecipe.TARGET_DATA])
            .reset_index(drop=True)
            .copy()
        )
        archive2 = (
            pd.concat([TestRecipe.ARCHIVE_DATA, TestRecipe.TARGET_DATA])
            .reset_index(drop=True)
            .copy()
        )

        messy_rp = permute_stitching_recipes(
            N_matches=2,
            matched_data=match_neighborhood(target2, archive2, tol=0.2),
            archive=archive2,
            testing=True,
        )

        # Test that there is no envelope collapse across ensemble members.
        # We define collapse across ensemble members as 'you don't have
        # realization 1 and realization 4 2070 getting matched to the same archive point.'
        # if r1-2050 and r4-2070 get matched to the same archive point,
        # that's fine. It's just multiple target  realizations having
        cols = [col for col in messy_rp if col.startswith("archive_")]

        for year in messy_rp["target_year"].unique():
            tmp = messy_rp[messy_rp["target_year"] == year].copy()
            unique_archive_rows = tmp[cols].drop_duplicates().copy()
            self.assertEqual(
                len(tmp), len(unique_archive_rows), "Envelope collapse has occurred"
            )
            del tmp
            del unique_archive_rows


if __name__ == "__main__":
    unittest.main()



"""Module for processing raw tas time series into 'chunks' of data used in the matching process."""

import math as math

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

import stitches.fx_util as util


def calculate_rolling_mean(data, size):
    """
    Calculate the rolling mean for the data frame with a user-defined size centered window.

    :param data: A data frame of the CMIP absolute temperature.
    :type data: pandas.core.frame.DataFrame
    :param size: An integer value for the size of the window to use when calculating the rolling mean.
    :type size: int
    :return: A pandas data frame of the smoothed time series with the rolling mean applied.
    """
    # Check inputs
    util.check_columns(
        data, {"model", "experiment", "ensemble", "year", "value", "variable"}
    )

    # Index the data frame by the year, so that the rolling mean respects the years
    data.index = data["year"]

    group_by = ["model", "experiment", "ensemble", "variable"]

    # previously, we just had a call:
    # rslt = data.groupby(group_by)['value'].rolling(size, center=True).mean().reset_index()
    # This returns the proper rolling mean on grouped data, but the first and last (size-1)/2
    # values in each grouped time series return as NaN because the full window isn't available
    # on both sides of the values there. We want to still have 'semi-smoothed' values there, as
    # in, we average whatever data we have in each grouped time series. So for a window size=5,
    # the smoothed 2099 value would be the average of 2097-2100 in our goal set up (as opposed
    # to 2097-2101 if we had the full 5 years of data centered on 2099).
    # In theory, pd.rolling(min_periods=1) does exactly what we want; however, it's weird on
    # grouped data. Sticking with the example of window size = 5, it was bringing in values from the
    # next group's time series to fill in 2099 and 2100.
    # To get pandas.rolling(min_periods-1) to apply correctly to grouped data, we have to wrap it
    # in a call to `transform` and save that as a new column:

    rslt = data.copy()
    rslt["rollingAvg"] = rslt.groupby(group_by)["value"].transform(
        lambda x: x.rolling(size, center=True, min_periods=1).mean()
    )
    # ^ seems like the way people usually get the functionality of a
    # dplyr group_by %>% mutate()  call.

    # rename so that value is the smoothed data:
    rslt = (
        rslt.drop(columns="value", axis=1)
        .rename(columns={"rollingAvg": "value"})
        .reset_index(drop=True)
    )

    return rslt


def chunk_ts(df, n, base_chunk=0):
    """
    Format a data frame into an array of data frames with n-sized years of successive data.

    This function takes a data frame of climate data and chunks it into separate periods,
    each containing data for a span of `n` years. It adds a 'chunk' column to the data frame
    to indicate the period each row belongs to.

    :param df: Data frame of climate data to chunk into different periods.
    :type df: pandas.DataFrame
    :param n: The size of the windows to chunk into separate periods.
    :type n: int
    :param base_chunk: A helper argument for creating staggered chunks, defaults to 0 (original behavior).
    :type base_chunk: int
    :return: A pandas DataFrame identical to `df` with the addition of a 'chunk' column.
    """
    # Check inputs
    df = df.drop_duplicates().reset_index(drop=True).copy()
    util.check_columns(df, {"year", "variable", "value"})
    if not (len(df["variable"].unique()) == 1):
        raise TypeError(f'Multiple variables discovered in "{df}"')

    # do the offset for the staggered windows
    if base_chunk > n:
        raise TypeError("base_chunk cannot be larger than n")
    df = df[base_chunk:].copy()

    # Add a column of data that categorizes the data into the different periods, right now we allow for
    # a list that is smaller than n at the end.
    # Save a copy of the length of the list
    yr_len = len(df["year"].unique())

    # Determine the number of unique n sized chunks, right now we assume that at most a chunk
    # can have n entries or less.
    n_chunks = math.ceil(yr_len / n)
    # Make an np array of the chunk labels, where each label repeats n times.
    chunk_labels = np.repeat(list(range(0, n_chunks)), n)
    # Subset the chunk list so that matches the length of the data frame.
    chunk_labels = chunk_labels[range(0, yr_len)]

    df["chunk"] = chunk_labels

    return df


def get_chunk_info(df):
    """
    Determine the value and the rate of change for each chunk.

    :param df: Data frame of climate data chunked into different periods.
    :type df: pandas.DataFrame
    :return: A pandas DataFrame with the chunk information, including the start and end years, the chunk value (fx),
             and the chunk rate of change (dx).
    """
    # Check the inputs
    flag = "index" in df.columns
    if flag:
        del df["index"]

    util.check_columns(df, {"year", "variable", "value", "chunk"})
    if not (len(df["variable"].unique()) == 1):
        raise TypeError(f'Multiple variables discovered in "{df}"')

    # Save information that will be added to
    extra_columns = list(set(df.columns).difference({"chunk", "value", "year"}))
    extra_info = df[extra_columns].drop_duplicates()
    if not (len(extra_info) == 1):
        raise TypeError("More than one type of data being read into the function.")

    # Group the data frame by the chunk label so that we can use a for loop
    # to extract information from each chunk of data.
    df_gby = df.groupby("chunk")

    # Use the for loop to work our way through the different chunks/periods of
    # data, gathering information about each start and stop year, the central
    # value, and the rate of change.
    index = 0
    for key, chunk in df_gby:
        # Save some information about the period of data, when it starts and stops.
        chunk["year"] = chunk["year"].astype(int)
        start_yr = min(chunk["year"])
        end_yr = max(chunk["year"])

        # Get the fx, the value of the center of the time period
        x = math.ceil(
            np.median(chunk["year"])
        )  # right now we select the one from the high year
        fx = chunk[chunk["year"] == x]["value"].values[0]

        # Prepare the x and y inputs from the data frame for the linear regression.
        # LinearRegression imported from sklearn.linear_model requires numpy arrays
        # instead of pandas objects.
        x_input = chunk["year"].values.reshape(-1, 1)
        y_input = chunk["value"].values.reshape(-1, 1)

        # Fit a linear regression of the period of time, extract the slope or dx
        # from the fit. Make sure that the dx is floating value that can be easily
        # stored in a data frame.
        model = LinearRegression()
        model.fit(x_input, y_input)
        dx = float(model.coef_[0])

        # Format the the chunk data into a pandas data frame.
        row = pd.DataFrame(
            [[start_yr, end_yr, x, fx, dx]],
            columns=["start_yr", "end_yr", "year", "fx", "dx"],
        )
        if index == 0:
            fx_dx_info = row
            index += 1
        else:
            fx_dx_info = pd.concat([fx_dx_info, row])

    # for loop should end here

    # Now add the extra or meta information to the data frame containing
    # the results. There is most likely a more efficient way to do this
    # in python but for now create an placeholder index column to join
    # the meta data (1 row of data) to the output data.
    extra_info["i"] = 1
    fx_dx_info["i"] = 1
    out = pd.merge(left=extra_info, right=fx_dx_info, left_on="i", right_on="i")
    out = out.loc[:, out.columns != "i"].copy()

    # Make sure that the objects being returned have int values for the
    # years instead of returning those values as a factor. This makes it
    # easier to work with data objects returned from other functions.
    data_types_dict = {"start_yr": "int32", "year": "int32", "end_yr": "int32"}
    out = out.astype(data_types_dict)

    return out


def subset_archive(staggered_archive, end_yr_vector):
    """
    Subset a staggered archive to entries with `end_yr` in `end_yr_vector`.

    This function takes a staggered archive with chunked data for a 9-year window
    following each year in 1850-2100 and subsets it to the entries with `end_yr`
    in `end_yr_vector`.

    :param staggered_archive: A formatted archive with chunked data starting in each year.
    :type staggered_archive: pandas.DataFrame
    :param end_yr_vector: Vector of end years to subset the archive to.
    :type end_yr_vector: list or similar iterable
    :return: A pandas DataFrame of the subsetted archive, same format but fewer entries.
    :rtype: pandas.DataFrame
    """
    out = (
        staggered_archive[staggered_archive["end_yr"].isin(end_yr_vector)]
        .reset_index(drop=True)
        .copy()
    )
    return out
